import logging
import os

from .. import entities, repositories, services

logger = logging.getLogger(name='dtlpy')


class PackageCodebaseType:
    ITEM = 'item'
    GIT = 'git'
    FILESYSTEM = 'filesystem'
    LOCAL = 'local'


def Codebase(**kwargs):
    """
    Factory function to init all codebases types
    """
    client_api = kwargs.pop('client_api', None)
    # take it out because we dont need it from the factory method
    _dict = kwargs.pop('_dict', None)

    if kwargs['type'] == PackageCodebaseType.GIT:
        cls = GitCodebase.from_json(_json=kwargs,
                                    client_api=client_api)
    elif kwargs['type'] == PackageCodebaseType.ITEM:
        cls = ItemCodebase.from_json(_json=kwargs,
                                     client_api=client_api)
    elif kwargs['type'] == PackageCodebaseType.FILESYSTEM:
        cls = FilesystemCodebase.from_json(_json=kwargs,
                                           client_api=client_api)
    elif kwargs['type'] == PackageCodebaseType.LOCAL:
        cls = LocalCodebase.from_json(_json=kwargs,
                                      client_api=client_api)
    else:
        raise ValueError('[Codebase constructor] Unknown codebase type: {}'.format(kwargs['type']))
    return cls


class GitCodebase(entities.DlEntity):
    type = entities.DlProperty(location=['type'], _type=str, default='git')
    git_url = entities.DlProperty(location=['gitUrl'], _type=str)
    git_tag = entities.DlProperty(location=['gitTag'], _type=str)
    credentials = entities.DlProperty(location=['credentials'], _type=dict)
    _codebases = None
    client_api: 'services.ClientApi'

    @property
    def is_remote(self):
        """ Return whether the codebase is managed remotely and supports upload-download"""
        return True

    @property
    def is_local(self):
        """ Return whether the codebase is locally and has no management implementations"""
        return not self.is_remote

    @property
    def codebases(self):
        if self._codebases is None:
            self._codebases = repositories.Codebases(client_api=self.client_api,
                                                     dataset=None)
        assert isinstance(self._codebases, repositories.Codebases)
        return self._codebases

    def to_json(self) -> dict:
        _dict = self._dict.copy()
        return _dict

    @classmethod
    def from_json(cls, _json: dict, client_api):
        """
        :param _json: platform json
        :param client_api: ApiClient entity
        """
        return cls(_dict=_json.copy(),
                   client_api=client_api)

    @property
    def git_user_name(self):
        return self.git_url.split('/')[-2]

    @property
    def git_repo_name(self):
        last = self.git_url.split('/')[-1]
        return os.path.splitext(last)[0]

    @property
    def git_username(self):
        if self.credentials is not None:
            return os.environ.get(
                self.credentials.get('username', {}).get('key', ''),
                None
            )
        return None

    @property
    def git_password(self):
        if self.credentials is not None:
            return os.environ.get(
                self.credentials.get('password', {}).get('key', ''),
                None
            )
        return None

    @staticmethod
    def is_git_repo(path):
        """
        :param path: `str` TODO: Currently only for local folder
        :return: `bool` testing if the path is valid git repo
        """
        return os.path.isdir(os.path.join(path, '.git'))

    def unpack(self, local_path):
        """
        Clones the git codebase
        :param local_path:
        """
        return self.codebases.clone_git(
            codebase=self,
            local_path=local_path
        )


class LocalCodebase(entities.DlEntity):
    type: str
    local_path: str
    _client_api: 'services.ClientApi'

    @property
    def is_remote(self):
        """ Return whether the codebase is managed remotely and supports upload-download"""
        return False

    @property
    def is_local(self):
        """ Return whether the codebase is locally and has no management implementations"""
        return not self.is_remote

    def to_json(self):
        _json = {'type': self.type,
                 'localPath': self._local_path}
        return _json

    @classmethod
    def from_json(cls, _json: dict, client_api):
        """
        :param _json: platform json
        :param client_api: ApiClient entity
        """
        return cls(
            client_api=client_api,
            local_path=_json.get('localPath', None),
            type=_json.get('type', None)
        )


class FilesystemCodebase(entities.DlEntity):
    type: str
    host_path: str
    container_path: str
    _client_api: 'services.ClientApi'

    @property
    def is_remote(self):
        """ Return whether the codebase is managed remotely and supports upload-download"""
        return False

    @property
    def is_local(self):
        """ Return whether the codebase is locally and has no management implementations"""
        return not self.is_remote

    def to_json(self):
        _json = super().to_json()
        if self.host_path is not None:
            _json['hostPath'] = self.host_path
        if self.container_path is not None:
            _json['containerPath'] = self.container_path
        return _json

    @classmethod
    def from_json(cls, _json: dict, client_api):
        """
        :param _json: platform json
        :param client_api: ApiClient entity
        """
        return cls(
            client_api=client_api,
            container_path=_json.get('containerPath', None),
            host_path=_json.get('hostPath', None),
            type=_json.get('type', None)
        )


class ItemCodebase(entities.DlEntity):
    type = entities.DlProperty(location=['type'], _type=str)
    item_id = entities.DlProperty(location=['itemId'], _type=str)
    _item = entities.DlProperty(location=['item'], _type=str, default=None)
    _codebases = None
    client_api: 'services.ClientApi'

    @property
    def item(self):
        if self._item is None:
            self._item = self.codebases.items_repository.get(item_id=self.item_id)
        return self._item

    @property
    def is_remote(self):
        """ Return whether the codebase is managed remotely and supports upload-download"""
        return True

    @property
    def is_local(self):
        """ Return whether the codebase is locally and has no management implementations"""
        return not self.is_remote

    @property
    def codebases(self):
        if self._codebases is None:
            if self._item is not None:
                dataset = self.item.dataset
            else:
                dataset = None
            self._codebases = repositories.Codebases(client_api=self.client_api,
                                                     dataset=dataset)
        assert isinstance(self._codebases, repositories.Codebases)
        return self._codebases

    def to_json(self) -> dict:
        _dict = self._dict.copy()
        _dict.pop('item', None)
        return _dict

    @classmethod
    def from_json(cls, _json: dict, client_api):
        """
        :param _json: platform json
        :param client_api: ApiClient entity
        """
        return cls(_dict=_json.copy(),
                   client_api=client_api)

    def unpack(self, local_path):
        """
        Clones the git codebase
        :param local_path:
        """
        return self.codebases.unpack(
            codebase_id=self.item_id,
            local_path=local_path
        )

    @property
    def version(self):
        return str(self.item.name.split('.')[0])

    @property
    def md5(self):
        md5 = None
        if 'system' in self.item.metadata:
            md5 = self.item.metadata['system'].get('md5', None)
        return md5

    @md5.setter
    def md5(self, md5):
        if 'system' not in self.item.metadata:
            self.item.metadata['system'] = dict()
        self.item.metadata['system']['md5'] = md5

    @property
    def description(self):
        description = None
        if 'system' in self.item.metadata:
            description = self.item.metadata['system'].get('description', None)
        return description

    @description.setter
    def description(self, description):
        if 'system' not in self.item.metadata:
            self.item.metadata['system'] = dict()
        self.item.metadata['system']['description'] = description

    def list_versions(self):
        """
        List Codebase versions
        """
        # get codebase name
        codebase_name = self.item.filename.split('/')[len(self.item.filename.split('/')) - 2]
        return self.codebases.list_versions(codebase_name=codebase_name)



def main(project_name, dataset_name, item_path):
    """
    Add any metadata to item
    :return:
    """
    # import Dataloop SDK
    import dtlpy as dl

    # get dataset
    dataset = dl.projects.get(project_name=project_name).datasets.get(dataset_name=dataset_name)

    # upload and claim item
    item = dataset.items.upload(local_path=item_path)

    # modify metadata
    item.metadata['user'] = dict()
    item.metadata['user']['MyKey'] = 'MyVal'
    # update and reclaim item
    item = item.update()

    # item in platform should have section 'user' in metadata with field 'MyKey' and value 'MyVal'



import attr
import logging
import traceback
from enum import Enum
from collections import namedtuple

from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class ExecutionStatus(str, Enum):
    SUCCESS = "success"
    FAILED = "failed"
    IN_PROGRESS = "in-progress"
    CREATED = "created"
    TERMINATED = 'terminated',
    ABORTED = 'aborted'
    CANCELED = 'canceled'
    SYSTEM_FAILURE = 'system-failure'


@attr.s
class Execution(entities.BaseEntity):
    """
    Service execution entity
    """
    # platform
    id = attr.ib()
    url = attr.ib(repr=False)
    creator = attr.ib()
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    input = attr.ib()
    output = attr.ib(repr=False)
    feedback_queue = attr.ib(repr=False)
    _status = attr.ib(repr=False)
    status_log = attr.ib(repr=False)
    sync_reply_to = attr.ib(repr=False)
    latest_status = attr.ib()
    function_name = attr.ib()
    duration = attr.ib()
    attempts = attr.ib()
    max_attempts = attr.ib()
    to_terminate = attr.ib(type=bool)

    # name changed
    trigger_id = attr.ib()
    service_id = attr.ib()
    project_id = attr.ib()
    service_version = attr.ib()
    package_id = attr.ib()
    package_name = attr.ib()
    package_revision = attr.ib()

    # sdk
    _client_api = attr.ib(type=ApiClient, repr=False)
    _service = attr.ib(repr=False)
    _project = attr.ib(repr=False, default=None)
    _repositories = attr.ib(repr=False)

    # optional
    pipeline = attr.ib(type=dict, default=None, repr=False)
    model = attr.ib(type=dict, default=None, repr=False)
    app = attr.ib(default=None)
    driver_id = attr.ib(default=None)

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['executions', 'services'])

        if self._project is not None:
            services_repo = self._project.services
            executions_repo = self._project.executions
        elif self._service is not None:
            services_repo = self._service.services
            executions_repo = self._service.executions
        else:
            services_repo = repositories.Services(client_api=self._client_api,
                                                  project=self._project,
                                                  package=None)
            executions_repo = repositories.Executions(client_api=self._client_api,
                                                      project=self._project,
                                                      service=self._service)

        r = reps(executions=executions_repo,
                 services=services_repo)
        return r

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @property
    def syncReplyTo(self):
        return self.sync_reply_to

    @property
    def feedbackQueue(self):
        return self.feedback_queue

    @property
    def status(self):
        return self._status

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    @property
    def executions(self):
        assert isinstance(self._repositories.executions, repositories.Executions)
        return self._repositories.executions

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, service=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json: platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            execution = Execution.from_json(_json=_json,
                                            client_api=client_api,
                                            project=None,
                                            service=service,
                                            is_fetched=is_fetched)
            status = True
        except Exception:
            execution = traceback.format_exc()
            status = False
        return status, execution

    @classmethod
    def from_json(cls, _json, client_api, project=None, service=None, is_fetched=True):
        """
        :param dict _json: platform json
        :param dl.ApiClient client_api: ApiClient entity
        :param dtlpy.entities.project.Project project: project entity
        :param dtlpy.entities.service.Service service:
        :param is_fetched: is Entity fetched from Platform
        """
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Execution has been fetched from a project that is not belong to it')
                project = None

        if service is not None:
            if service.id != _json.get('serviceId', None):
                logger.warning('Execution has been fetched from a service that is not belong to it')
                service = None

        inst = cls(
            feedback_queue=_json.get('feedbackQueue', None),
            service_id=_json.get('serviceId', None),
            project_id=_json.get('projectId', None),
            latest_status=_json.get('latestStatus', None),
            sync_reply_to=_json.get('syncReplyTo', None),
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            creator=_json.get('creator', None),
            trigger_id=_json.get('triggerId', None),
            attempts=_json.get('attempts', None),
            max_attempts=_json.get('maxAttempts', None),
            output=_json.get('output', None),
            status=_json.get('status', None),
            status_log=_json.get('statusLog', None),
            duration=_json.get('duration', None),
            function_name=_json.get('functionName', entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME),
            input=_json.get('input', None),
            url=_json.get('url', None),
            id=_json.get('id', None),
            to_terminate=_json.get('toTerminate', False),
            client_api=client_api,
            project=project,
            service=service,
            service_version=_json.get('serviceVersion', False),
            package_id=_json.get('packageId', None),
            package_name=_json.get('packageName', None),
            pipeline=_json.get('pipeline', None),
            model=_json.get('model', None),
            package_revision=_json.get('packageRevision', None),
            app=_json.get('app', None),
            driver_id=_json.get('driverId', None)
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        # get excluded
        _json = attr.asdict(
            self, filter=attr.filters.exclude(
                attr.fields(Execution)._client_api,
                attr.fields(Execution)._service,
                attr.fields(Execution)._project,
                attr.fields(Execution).to_terminate,
                attr.fields(Execution)._repositories,
                attr.fields(Execution).project_id,
                attr.fields(Execution).service_id,
                attr.fields(Execution).trigger_id,
                attr.fields(Execution).function_name,
                attr.fields(Execution).max_attempts,
                attr.fields(Execution).latest_status,
                attr.fields(Execution).service_version,
                attr.fields(Execution).package_id,
                attr.fields(Execution).package_name,
                attr.fields(Execution).status_log,
                attr.fields(Execution)._status,
                attr.fields(Execution).created_at,
                attr.fields(Execution).updated_at,
                attr.fields(Execution).feedback_queue,
                attr.fields(Execution).sync_reply_to,
                attr.fields(Execution).pipeline,
                attr.fields(Execution).model,
                attr.fields(Execution).package_revision,
                attr.fields(Execution).driver_id,
            )
        )

        # rename
        _json['projectId'] = self.project_id
        _json['triggerId'] = self.trigger_id
        _json['serviceId'] = self.service_id
        _json['functionName'] = self.function_name
        _json['latestStatus'] = self.latest_status
        _json['maxAttempts'] = self.max_attempts
        _json['toTerminate'] = self.to_terminate
        _json['serviceVersion'] = self.service_version
        _json['packageId'] = self.package_id
        _json['packageName'] = self.package_name
        _json['statusLog'] = self.status_log
        _json['status'] = self._status
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        _json['feedbackQueue'] = self.feedback_queue
        _json['syncReplyTo '] = self.sync_reply_to
        _json['packageRevision'] = self.package_revision
        _json['driverId'] = self.driver_id

        if self.pipeline:
            _json['pipeline'] = self.pipeline
        if self.model:
            _json['model'] = self.model

        return _json

    @property
    def pipeline_id(self) -> str:
        pipeline_id = None
        if self.pipeline:
            pipeline_id = self.pipeline.get('id', None)
        return pipeline_id

    @property
    def node_id(self) -> str:
        node_id = None
        if self.pipeline:
            node_id = self.pipeline.get('nodeId', None)
        return node_id

    @property
    def pipeline_execution_id(self) -> str:
        pipeline_execution_id = None
        if self.pipeline:
            pipeline_execution_id = self.pipeline.get('executionId', None)
        return pipeline_execution_id

    @property
    def service(self):
        if self._service is None:
            self._service = self.services.get(service_id=self.service_id, fetch=None)
        assert isinstance(self._service, entities.Service)
        return self._service

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    def get_latest_status(self):
        self.latest_status = self.executions.get(execution_id=self.id).latest_status
        return self.latest_status

    @service.setter
    def service(self, service):
        if not isinstance(service, entities.Service):
            raise ValueError('Must input a valid service entity')
        self._service = service

    def progress_update(
            self,
            status: ExecutionStatus = None,
            percent_complete: int = None,
            message: str = None,
            output: str = None,
            service_version: str = None
    ):
        """
        Update Execution Progress

        :param str status: ExecutionStatus
        :param int percent_complete: percent complete
        :param str message: message to update the progress state
        :param str output: output
        :param str service_version: service version
        :return: Service execution object
        """
        return self.executions.progress_update(
            execution_id=self.id,
            status=status,
            percent_complete=percent_complete,
            message=message,
            output=output,
            service_version=service_version
        )

    def update(self):
        """
        Update execution changes to platform

        :return: execution entity
        """
        return self.executions.update(execution=self)

    def logs(self, follow=False, log_level='DEBUG'):
        """
        Print logs for execution

        :param follow: keep stream future logs
        :param str log_level: the log level to display
        """
        self.services.log(execution_id=self.id,
                          view=True,
                          service=self.service,
                          follow=follow,
                          start=self.created_at,
                          log_level=log_level)

    def increment(self):
        """
        Increment attempts

        :return:
        """
        self.attempts = self.executions.increment(execution=self)

    def rerun(self, sync: bool = False):
        """
        Re-run

        :return: Execution object
        """
        return self.executions.rerun(execution=self, sync=sync)

    def terminate(self):
        """
        Terminate execution

        :return: execution object
        """
        return self.executions.terminate(execution=self)

    def wait(self):
        """
        Wait for execution

        :return: Service execution object
        """
        return self.executions.wait(execution_id=self.id)

    def in_progress(self):
        return self.latest_status['status'] not in [ExecutionStatus.FAILED,
                                                    ExecutionStatus.SUCCESS,
                                                    ExecutionStatus.TERMINATED,
                                                    ExecutionStatus.ABORTED,
                                                    ExecutionStatus.CANCELED,
                                                    ExecutionStatus.SYSTEM_FAILURE]



import logging
import pandas as pd
from dtlpy import entities, exceptions
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Analytics:
    """
    Time series Repository
    """

    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.times_series repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ############
    #  methods #
    ############
    def get_samples(self, query=None, return_field: str = None, return_raw: bool = False) -> pd.DataFrame:
        """
        Get Analytics table

        :param dict query: match filters to get specific data from series
        :param str return_field: name of field to return from response. default: "samples"
        :param bool return_raw: return the response with out converting
        :return: Analytics table
        :rtype: pd.DataFrame
        """
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/analytics/query',
                                                         json_req=query)
        if success:
            if return_field is not None:
                res = response.json()[return_field]
            else:
                res = response.json()
            if return_raw:
                return res
            if isinstance(res, dict):
                df = pd.DataFrame.from_dict(res, orient="index")
            elif isinstance(res, list):
                df = pd.DataFrame(res)
            else:
                raise ValueError('unknown return type for time series: {}'.format(type(res)))
        else:
            raise exceptions.PlatformException(response)
        return df

    def report_metrics(self, samples):
        """
        Report metrics

        :param samples: table samples to report
        :return: True/ False
        :rtype: bool
        """
        if not isinstance(samples, list):
            samples = [samples]

        samples = [s.to_json() if not isinstance(s, dict) else s for s in samples]

        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/analytics/metric',
                                                         json_req=samples)
        return success



from collections import namedtuple
from enum import Enum
import traceback
import logging
import attr

from .. import repositories, miscellaneous, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class MemberRole(str, Enum):
    OWNER = "owner"
    DEVELOPER = "engineer"
    ANNOTATOR = "annotator"
    ANNOTATION_MANAGER = "annotationManager"


@attr.s()
class Project(entities.BaseEntity):
    """
    Project entity
    """

    _contributors = attr.ib(repr=False)
    created_at = attr.ib()
    creator = attr.ib()
    id = attr.ib()
    url = attr.ib()
    name = attr.ib()
    org = attr.ib(repr=False)
    updated_at = attr.ib(repr=False)
    role = attr.ib(repr=False)
    account = attr.ib(repr=False)
    is_blocked = attr.ib(repr=False)
    archived = attr.ib(repr=False)

    # name change
    feature_constraints = attr.ib()

    # api
    _client_api = attr.ib(type=ApiClient, repr=False)

    # repositories
    _repositories = attr.ib(repr=False)

    @property
    def isBlocked(self):
        return self.is_blocked

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple(
            'repositories',
            'projects triggers datasets items recipes packages codebases artifacts times_series services '
            'executions assignments tasks bots webhooks models analytics ontologies '
            'drivers pipelines feature_sets features integrations settings apps dpks compositions'
        )
        datasets = repositories.Datasets(client_api=self._client_api, project=self)
        return reps(
            projects=repositories.Projects(client_api=self._client_api),
            webhooks=repositories.Webhooks(client_api=self._client_api, project=self),
            items=repositories.Items(client_api=self._client_api, datasets=datasets, project=self),
            recipes=repositories.Recipes(client_api=self._client_api, project=self, project_id=self.id),
            datasets=datasets,
            executions=repositories.Executions(client_api=self._client_api, project=self),
            triggers=repositories.Triggers(client_api=self._client_api, project=self),
            packages=repositories.Packages(project=self, client_api=self._client_api),
            models=repositories.Models(project=self, client_api=self._client_api),
            codebases=repositories.Codebases(project=self, client_api=self._client_api),
            artifacts=repositories.Artifacts(project=self, client_api=self._client_api),
            times_series=repositories.TimesSeries(project=self, client_api=self._client_api),
            services=repositories.Services(client_api=self._client_api, project=self),
            assignments=repositories.Assignments(project=self, client_api=self._client_api),
            tasks=repositories.Tasks(client_api=self._client_api, project=self),
            bots=repositories.Bots(client_api=self._client_api, project=self),
            analytics=repositories.Analytics(client_api=self._client_api, project=self),
            ontologies=repositories.Ontologies(client_api=self._client_api, project=self),
            drivers=repositories.Drivers(client_api=self._client_api, project=self),
            pipelines=repositories.Pipelines(client_api=self._client_api, project=self),
            feature_sets=repositories.FeatureSets(client_api=self._client_api, project=self),
            features=repositories.Features(client_api=self._client_api, project=self),
            integrations=repositories.Integrations(client_api=self._client_api, project=self),
            settings=repositories.Settings(client_api=self._client_api,
                                           project=self,
                                           resource=self,
                                           resource_type=entities.PlatformEntityType.PROJECT),
            apps=repositories.Apps(client_api=self._client_api, project=self),
            dpks=repositories.Dpks(client_api=self._client_api, project=self),
            compositions=repositories.Compositions(client_api=self._client_api, project=self)
        )

    @property
    def drivers(self):
        assert isinstance(self._repositories.drivers, repositories.Drivers)
        return self._repositories.drivers

    @property
    def integrations(self):
        assert isinstance(self._repositories.integrations, repositories.Integrations)
        return self._repositories.integrations

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}".format(self.id))

    @property
    def triggers(self):
        assert isinstance(self._repositories.triggers, repositories.Triggers)
        return self._repositories.triggers

    @property
    def ontologies(self):
        assert isinstance(self._repositories.ontologies, repositories.Ontologies)
        return self._repositories.ontologies

    @property
    def items(self):
        assert isinstance(self._repositories.items, repositories.Items)
        return self._repositories.items

    @property
    def recipes(self):
        assert isinstance(self._repositories.recipes, repositories.Recipes)
        return self._repositories.recipes

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    @property
    def executions(self):
        assert isinstance(self._repositories.executions, repositories.Executions)
        return self._repositories.executions

    @property
    def apps(self):
        assert isinstance(self._repositories.apps, repositories.Apps)
        return self._repositories.apps

    @property
    def dpks(self):
        assert isinstance(self._repositories.dpks, repositories.Dpks)
        return self._repositories.dpks

    @property
    def compositions(self):
        assert isinstance(self._repositories.compositions, repositories.Compositions)
        return self._repositories.compositions

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def datasets(self):
        assert isinstance(self._repositories.datasets, repositories.Datasets)
        return self._repositories.datasets

    @property
    def pipelines(self):
        assert isinstance(self._repositories.pipelines, repositories.Pipelines)
        return self._repositories.pipelines

    @property
    def packages(self):
        assert isinstance(self._repositories.packages, repositories.Packages)
        return self._repositories.packages

    @property
    def models(self):
        assert isinstance(self._repositories.models, repositories.Models)
        return self._repositories.models

    @property
    def codebases(self):
        assert isinstance(self._repositories.codebases, repositories.Codebases)
        return self._repositories.codebases

    @property
    def webhooks(self):
        assert isinstance(self._repositories.webhooks, repositories.Webhooks)
        return self._repositories.webhooks

    @property
    def artifacts(self):
        assert isinstance(self._repositories.artifacts, repositories.Artifacts)
        return self._repositories.artifacts

    @property
    def times_series(self):
        assert isinstance(self._repositories.times_series, repositories.TimesSeries)
        return self._repositories.times_series

    @property
    def assignments(self):
        assert isinstance(self._repositories.assignments, repositories.Assignments)
        return self._repositories.assignments

    @property
    def tasks(self):
        assert isinstance(self._repositories.tasks, repositories.Tasks)
        return self._repositories.tasks

    @property
    def bots(self):
        assert isinstance(self._repositories.bots, repositories.Bots)
        return self._repositories.bots

    @property
    def analytics(self):
        assert isinstance(self._repositories.analytics, repositories.Analytics)
        return self._repositories.analytics

    @property
    def feature_sets(self):
        assert isinstance(self._repositories.feature_sets, repositories.FeatureSets)
        return self._repositories.feature_sets

    @property
    def features(self):
        assert isinstance(self._repositories.features, repositories.Features)
        return self._repositories.features

    @property
    def settings(self):
        assert isinstance(self._repositories.settings, repositories.Settings)
        return self._repositories.settings

    @property
    def contributors(self):
        return miscellaneous.List([entities.User.from_json(_json=_json,
                                                           client_api=self._client_api,
                                                           project=self) for _json in self._contributors])

    @staticmethod
    def _protected_from_json(_json, client_api):
        """
        Same as from_json but with try-except to catch if error

        :param dict _json: platform json
        :param dl.ApiClient client_api: ApiClient entity
        :return:
        """
        try:
            project = Project.from_json(_json=_json,
                                        client_api=client_api)
            status = True
        except Exception:
            project = traceback.format_exc()
            status = False
        return status, project

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Project object from a json

        :param bool is_fetched: is Entity fetched from Platform
        :param dict _json: _json response from host
        :param dl.ApiClient client_api: ApiClient entity
        :return: Project object
        :rtype: dtlpy.entities.project.Project
        """
        inst = cls(feature_constraints=_json.get('featureConstraints', None),
                   contributors=_json.get('contributors', None),
                   is_blocked=_json.get('isBlocked', None),
                   created_at=_json.get('createdAt', None),
                   updated_at=_json.get('updatedAt', None),
                   creator=_json.get('creator', None),
                   account=_json.get('account', None),
                   name=_json.get('name', None),
                   role=_json.get('role', None),
                   org=_json.get('org', None),
                   id=_json.get('id', None),
                   url=_json.get('url', None),
                   archived=_json.get('archived', None),
                   client_api=client_api)
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of project object

        :return: platform json format of project object
        :rtype: dict
        """
        output_dict = attr.asdict(self,
                                  filter=attr.filters.exclude(attr.fields(Project)._client_api,
                                                              attr.fields(Project)._repositories,
                                                              attr.fields(Project).feature_constraints,
                                                              attr.fields(Project)._contributors,
                                                              attr.fields(Project).created_at,
                                                              attr.fields(Project).updated_at,
                                                              attr.fields(Project).is_blocked,
                                                              ))
        output_dict['contributors'] = self._contributors
        output_dict['featureConstraints'] = self.feature_constraints
        output_dict['createdAt'] = self.created_at
        output_dict['updatedAt'] = self.updated_at
        output_dict['isBlocked'] = self.is_blocked

        return output_dict

    def delete(self, sure=False, really=False):
        """
        Delete the project forever!

        :param bool sure: Are you sure you want to delete?
        :param bool really: Really really sure?
        :return: True if success, error if not
        :rtype: bool
        """
        return self.projects.delete(project_id=self.id,
                                    sure=sure,
                                    really=really)

    def update(self, system_metadata=False):
        """
        Update the project

        :param bool system_metadata: optional - True, if you want to change metadata system
        :return: Project object
        :rtype: dtlpy.entities.project.Project
        """
        return self.projects.update(project=self, system_metadata=system_metadata)

    def checkout(self):
        """
        Checkout (switch) to a project to work on.

        """
        self.projects.checkout(project=self)

    def open_in_web(self):
        """
        Open the project in web platform

        """
        self._client_api._open_in_web(url=self.platform_url)

    def add_member(self, email, role: MemberRole = MemberRole.DEVELOPER):
        """
        Add a member to the project.

        :param str email: member email
        :param role: The required role for the user. Use the enum dl.MemberRole
        :return: dict that represent the user
        :rtype: dict
        """
        return self.projects.add_member(email=email, role=role, project_id=self.id)

    def update_member(self, email, role: MemberRole = MemberRole.DEVELOPER):
        """
        Update member's information/details from the project.

        :param str email: member email
        :param role: The required role for the user. Use the enum dl.MemberRole
        :return: dict that represent the user
        :rtype: dict
        """
        return self.projects.update_member(email=email, role=role, project_id=self.id)

    def remove_member(self, email):
        """
        Remove a member from the project.

        :param str email: member email
        :return: dict that represents the user
        :rtype: dict
        """
        return self.projects.remove_member(email=email, project_id=self.id)

    def list_members(self, role: MemberRole = None):
        """
        List the project members.

        :param role: The required role for the user. Use the enum dl.MemberRole
        :return: list of the project members
        :rtype: list
        """
        return self.projects.list_members(project=self, role=role)



def main():
    import dtlpy as dl
    import matplotlib.pyplot as plt

    # Get project and dataset
    project = dl.projects.get(project_name='Food')
    dataset = project.datasets.get(dataset_name='BeansDataset')
    item = dataset.items.get(filepath='/flying boxes.mp4')

    ############################
    # using annotation builder #
    ############################
    # create annotation builder
    builder = item.annotations.builder()

    for i_frame in range(100):
        # go over 100 frame
        for i_detection in range(10):
            # for each frame we have 10 different detections (location is just for the example)
            builder.add(annotation_definition=dl.Box(top=2 * i_frame,
                                                     left=2 * i_detection,
                                                     bottom=2 * i_frame + 10,
                                                     right=2 * i_detection + 100,
                                                     label="moving box"),
                        frame_num=i_frame,  # set the frame for the annotation
                        object_id=i_detection + 1
                        # need to input the element id to create the connection between frames
                        )
            # starting from frame 50 add another 10 new annotations of a moving point
            if i_frame > 50:
                builder.add(annotation_definition=dl.Point(x=2 * i_frame,
                                                           y=2 * i_detection,
                                                           label="moving point"),
                            frame_num=i_frame,
                            object_id=20 + (i_detection + 1))
    # get frame annotations
    frame_annotations = builder.get_frame(frame_num=55)
    # Plot the annotations in frame 55 of the created annotations
    plt.figure()
    plt.imshow(frame_annotations.show())

    # plot each annotations separately
    for annotation in frame_annotations:
        plt.figure()
        plt.imshow(annotation.show())
        plt.title(annotation.label)

    # Add the annotations to platform
    item.annotations.upload(builder)

    #####################
    # single annotation #
    #####################
    # create one annotations for a video (without using tracker and object ids)
    annotation = dl.Annotation.new(item=item)

    for i_frame in range(100):
        # go over 100 frame
        annotation.add_frame(annotation_definition=dl.Box(top=2 * i_frame,
                                                          left=2 * (i_frame + 10),
                                                          bottom=2 * (i_frame + 50),
                                                          right=2 * (i_frame + 100),
                                                          label="moving box",
                                                          ),
                             frame_num=i_frame,  # set the frame for the annotation
                             )

    # upload to platform
    annotation.upload()

    ##############################################
    # show annotation state in a specified frame #
    ##############################################

    # Get from platform
    annotations = item.annotations.list()

    # Plot the annotations in frame 55 of the created annotations
    plt.figure()
    plt.imshow(annotations.get_frame(frame_num=55).show())

    # Play video with the Dataloop video player
    annotations.video_player()



import numpy as np
import logging

from . import BaseAnnotationDefinition

logger = logging.getLogger(name='dtlpy')


class Polygon(BaseAnnotationDefinition):
    """
    Polygon annotation object
    """
    type = "segment"

    def __init__(self, geo, label, attributes=None, description=None):
        super().__init__(description=description, attributes=attributes)
        self.geo = geo
        self.label = label

    @property
    def x(self):
        return self.geo[:, 0]

    @property
    def y(self):
        return self.geo[:, 1]

    @property
    def left(self):
        return np.min(self.x)

    @property
    def top(self):
        return np.min(self.y)

    @property
    def right(self):
        return np.max(self.x)

    @property
    def bottom(self):
        return np.max(self.y)

    def to_coordinates(self, color):
        return [[{"x": float(x), "y": float(y)} for x, y in self.geo]]

    @staticmethod
    def from_coordinates(coordinates):
        return np.asarray([[pt["x"], pt["y"]] for pt in coordinates])

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        if thickness is None:
            thickness = 2

        # create image to draw on
        if alpha != 1:
            overlay = image.copy()
        else:
            overlay = image

        overlay = cv2.drawContours(
            image=overlay,
            contours=[np.round(self.geo).astype(int)],
            contourIdx=-1,
            color=color,
            thickness=thickness,
        )

        if not isinstance(color, int) and len(color) == 4 and color[3] != 255:
            # add with opacity
            image = cv2.addWeighted(src1=overlay,
                                    alpha=alpha,
                                    src2=image,
                                    beta=1 - alpha,
                                    gamma=0)
        else:
            image = overlay

        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    @classmethod
    def from_segmentation(cls, mask, label, attributes=None, epsilon=None, max_instances=1, min_area=0):
        """
        Convert binary mask to Polygon

        :param mask: binary mask (0,1)
        :param label: annotation label
        :param attributes: annotations list of attributes
        :param epsilon: from opencv: specifying the approximation accuracy. This is the maximum distance between the original curve and its approximation. if 0 all points are returns
        :param max_instances: number of max instances to return. if None all wil be returned
        :param min_area: remove polygons with area lower thn this threshold (pixels)

        :return: Polygon annotation
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            raise ImportError('opencv not found. Must install to perform this function')

        # mask float
        mask = 1. * mask
        # normalize to 1
        mask /= np.max(mask)
        # threshold the mask
        ret, thresh = cv2.threshold(mask, 0.5, 255, 0)
        # find contours
        major, minor, _ = cv2.__version__.split(".")
        if int(major) > 3:
            contours, hierarchy = cv2.findContours(thresh.astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
        else:
            _, contours, hierarchy = cv2.findContours(thresh.astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

        if len(contours) == 0:
            # no contours were found
            new_pts_list = []
        else:
            # calculate contours area
            areas = np.asarray([cv2.contourArea(cnt) for cnt in contours])
            # take onr contour with maximum area
            sorted_areas_inds = areas.argsort()[::-1]
            filtered_contours = [contours[s_ind] for s_ind in
                                 sorted_areas_inds if areas[s_ind] > min_area]  # filter by area size
            filtered_contours = filtered_contours[:max_instances]  # take only the first max_instance of the results
            # estimate contour to reduce number of points
            new_pts_list = list()
            for curve in filtered_contours:
                if epsilon is None:
                    epsilon = 0.0005 * cv2.arcLength(curve=curve,
                                                     closed=True)
                estimated_polygon = np.squeeze(cv2.approxPolyDP(curve=curve,
                                                                epsilon=epsilon,
                                                                closed=True))
                if len(estimated_polygon.shape) == 1:
                    new_pts_list.append(np.squeeze(curve))
                else:
                    new_pts_list.append(estimated_polygon)
        polygons = [cls(geo=new_pts,
                        label=label,
                        attributes=attributes,
                        ) for new_pts in new_pts_list]

        if len(polygons) == 1:
            polygons = polygons[0]
        return polygons

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            geo = cls.from_coordinates(coordinates=_json["coordinates"][0])
        elif "data" in _json:
            geo = cls.from_coordinates(coordinates=_json["data"][0])
        else:
            raise ValueError(
                'can not find "coordinates" or "data" in annotation. id: %s'
                % _json["id"]
            )
        return cls(
            geo=geo,
            label=_json["label"],
            attributes=_json.get("attributes", None),
        )



import traceback
import logging
import attr
from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class NotificationEventContext:

    def __init__(self, context: dict):
        self.project: str = context.get('project', None)
        self.org: str = context.get('org', None)
        self.pipeline: str = context.get('pipeline', None)
        self.service: str = context.get('service', None)
        self.node: str = context.get('node', None)

    @classmethod
    def from_json(cls, _json: dict):
        return cls(context=_json)

    def to_json(self):
        _json = dict()
        if self.project:
            _json['project'] = self.project
        if self.org:
            _json['org'] = self.org
        if self.pipeline:
            _json['pipeline'] = self.pipeline
        if self.service:
            _json['service'] = self.service
        if self.node:
            _json['node'] = self.node
        return _json


@attr.s
class Message(entities.BaseEntity):
    """
    Message object
    """
    title = attr.ib(type=str)
    description = attr.ib(type=str)
    context = attr.ib(type=NotificationEventContext)
    read = attr.ib(type=int)
    dismissed = attr.ib(type=int)
    new = attr.ib(type=int)

    # from camel case to snake case
    user_id = attr.ib(type=str)
    notification_id = attr.ib(type=str)
    resource_action = attr.ib(type=str)
    notification_code = attr.ib(type=str)
    resource_type = attr.ib(type=str)
    resource_id = attr.ib(type=str)
    resource_name = attr.ib(type=str)

    # api
    _client_api = attr.ib(type=ApiClient, repr=False)

    # entities
    _project = attr.ib(default=None, repr=False)

    @staticmethod
    def _protected_from_json(
            _json: dict,
            client_api: ApiClient,
            project: entities.Project = None,
            is_fetched=True
    ):
        """
        Same as from_json but with try-except to catch if error

        :param project: message's project
        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return: Message object
        """
        try:
            message = Message.from_json(project=project,
                                        _json=_json,
                                        client_api=client_api,
                                        is_fetched=is_fetched)
            status = True
        except Exception:
            message = traceback.format_exc()
            status = False
        return status, message

    @classmethod
    def from_json(cls,
                  _json: dict,
                  client_api: ApiClient,
                  project: entities.Project = None,
                  is_fetched=True
                  ):
        """
        Build a Message entity object from a json

        :param project: message's project
        :param dict _json: _json response from host
        :param client_api: ApiClient entityÏ
        :param bool is_fetched: is Entity fetched from Platform
        :return: Dataset object
        :rtype: dtlpy.entities.message.Message
        """

        inst = cls(
            title=_json.get('title', None),
            description=_json.get('description', None),
            context=NotificationEventContext.from_json(_json=_json.get('context', None)),
            read=_json.get('read', None),
            dismissed=_json.get('dismissed', None),
            new=_json.get('new', None),
            user_id=_json.get('userId', None),
            notification_id=_json.get('notificationId', None),
            resource_action=_json.get('resourceAction', None),
            notification_code=_json.get('notificationCode', None),
            resource_type=_json.get('resourceType', None),
            resource_id=_json.get('resourceId', None),
            resource_name=_json.get('resourceName', None),
            client_api=client_api,
            project=project
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(
            self, filter=attr.filters.exclude(
                attr.fields(Message)._client_api,
                attr.fields(Message)._project,
                attr.fields(Message).is_fetched,
                attr.fields(Message).context,
                attr.fields(Message).user_id,
                attr.fields(Message).notification_id,
                attr.fields(Message).resource_action,
                attr.fields(Message).notification_code,
                attr.fields(Message).resource_type,
                attr.fields(Message).resource_id,
                attr.fields(Message).resource_name
            )
        )

        _json['context'] = self.context.to_json()
        _json['userId'] = self.user_id
        _json['notificationId'] = self.notification_id
        _json['resourceAction'] = self.resource_action
        _json['notificationCode'] = self.notification_code
        _json['resourceType'] = self.resource_type
        _json['resourceId'] = self.resource_id
        _json['resourceName'] = self.resource_name

        return _json

    @property
    def project(self):
        if self._project is None and self.context.project is not None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.context.project)
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project



# from https://pypi.org/project/aiohttp-retry/ with modification to python 3.5.4
import asyncio
import logging
from aiohttp import ClientSession, ClientResponse
from typing import Any, Callable, Optional, Set, Type

# Options
_RETRY_ATTEMPTS = 3
_RETRY_START_TIMEOUT = 0.1
_RETRY_MAX_TIMEOUT = 30
_RETRY_FACTOR = 2


class _RequestContext:
    def __init__(self, request: Callable[..., Any],  # Request operation, like POST or GET
                 url: str,  # Just url
                 retry_attempts: int = _RETRY_ATTEMPTS,  # How many times we should retry
                 retry_start_timeout: float = _RETRY_START_TIMEOUT,  # Base timeout time, then it exponentially grow
                 retry_max_timeout: float = _RETRY_MAX_TIMEOUT,  # Max possible timeout between tries
                 retry_factor: float = _RETRY_FACTOR,  # How much we increase timeout each time
                 retry_for_statuses: Optional[Set[int]] = None,  # On which statuses we should retry
                 retry_exceptions: Optional[Set[Type]] = None,  # On which exceptions we should retry
                 **kwargs: Any
                 ) -> None:
        self._request = request
        self._url = url

        self._retry_attempts = retry_attempts
        self._retry_start_timeout = retry_start_timeout
        self._retry_max_timeout = retry_max_timeout
        self._retry_factor = retry_factor

        if retry_for_statuses is None:
            retry_for_statuses = set()
        self._retry_for_statuses = retry_for_statuses

        if retry_exceptions is None:
            retry_exceptions = set()
        self._retry_exceptions = retry_exceptions

        self._kwargs = kwargs

        self._current_attempt = 0
        self._response = None

    def _exponential_timeout(self) -> float:
        timeout = self._retry_start_timeout * (self._retry_factor ** (self._current_attempt - 1))
        return min(timeout, self._retry_max_timeout)

    def _check_code(self, code: int) -> bool:
        return 500 <= code <= 599 or code in self._retry_for_statuses

    async def _do_request(self) -> ClientResponse:
        try:
            self._current_attempt += 1
            response = await self._request(url=self._url, **self._kwargs)
            code = response.status
            if self._current_attempt < self._retry_attempts and self._check_code(code):
                retry_wait = self._exponential_timeout()
                await asyncio.sleep(retry_wait)
                return await self._do_request()
            self._response = response
            return response

        except Exception as e:
            retry_wait = self._exponential_timeout()
            if self._current_attempt < self._retry_attempts:
                for exc in self._retry_exceptions:
                    if isinstance(e, exc):
                        await asyncio.sleep(retry_wait)
                        return await self._do_request()

            raise e

    async def __aenter__(self) -> ClientResponse:
        return await self._do_request()

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self._response is not None:
            if not self._response.closed:
                self._response.close()


class RetryClient:
    def __init__(self, logger: Any = None, *args: Any, **kwargs: Any) -> None:
        self._client = ClientSession(*args, **kwargs)
        self._closed = False

        if logger is None:
            logger = logging.getLogger("aiohttp_retry")

        self._logger = logger

    def __del__(self) -> None:
        if not self._closed:
            self._logger.warning("Aiohttp retry client was not closed")

    @staticmethod
    def _request(request: Callable[..., Any], url: str, **kwargs: Any) -> _RequestContext:
        return _RequestContext(request, url, **kwargs)

    def get(self, url: str, **kwargs: Any) -> _RequestContext:
        return self._request(self._client.get, url, **kwargs)

    def options(self, url: str, **kwargs: Any) -> _RequestContext:
        return self._request(self._client.options, url, **kwargs)

    def head(self, url: str, **kwargs: Any) -> _RequestContext:
        return self._request(self._client.head, url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> _RequestContext:
        return self._request(self._client.post, url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> _RequestContext:
        return self._request(self._client.put, url, **kwargs)

    def patch(self, url: str, **kwargs: Any) -> _RequestContext:
        return self._request(self._client.patch, url, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> _RequestContext:
        return self._request(self._client.delete, url, **kwargs)

    async def close(self) -> None:
        await self._client.close()
        self._closed = True

    async def __aenter__(self) -> 'RetryClient':
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()



import logging

import attr
import typing
from enum import Enum
from .. import entities

logger = logging.getLogger(name='dtlpy')


class SlotPostActionType(str, Enum):
    DOWNLOAD = 'download'
    DRAW_ANNOTATION = 'drawAnnotation'
    NO_ACTION = 'noAction'


class SlotDisplayScopeResource(str, Enum):
    ANNOTATION = 'annotation'
    ITEM = 'item'
    DATASET = 'dataset'
    DATASET_QUERY = 'datasetQuery'
    TASK = 'task'


class UiBindingPanel(str, Enum):
    BROWSER = "browser",
    STUDIO = "studio",
    TABLE = "table",
    ALL = "all"


class SlotPostAction(entities.DlEntity):
    type: str = entities.DlProperty(location=['type'],
                                    _type=str,
                                    default=SlotPostActionType.NO_ACTION)

    @classmethod
    def from_json(cls, _json):
        inst = cls(_dict=_json)
        return inst

    def to_json(self):
        _json = self._dict.copy()
        return _json


class SlotDisplayScope(entities.DlEntity):
    resource: str = entities.DlProperty(location=['resource'],
                                        _type=str)
    filters: entities.Filters = entities.DlProperty(location=['filter'],
                                                    _type=entities.Filters)
    panel: str = entities.DlProperty(location=['panel'],
                                     _type=str,
                                     default=UiBindingPanel.ALL)

    @classmethod
    def from_json(cls, _json):
        inst = cls(_dict=_json)
        return inst

    @staticmethod
    def get_resource(resource: str):
        if resource in [SlotDisplayScopeResource.DATASET, SlotDisplayScopeResource.DATASET_QUERY]:
            return entities.FiltersResource.DATASET
        elif resource == SlotDisplayScopeResource.ITEM:
            return entities.FiltersResource.ITEM
        elif resource == SlotDisplayScopeResource.ANNOTATION:
            return entities.FiltersResource.ANNOTATION

    def to_json(self):
        _json = self._dict.copy()
        if isinstance(self.filters, entities.Filters):
            _json['filter'] = self.filters.prepare(query_only=True)['filter']
        elif isinstance(self.filters, dict):
            _json['filter'] = self.filters
        return _json


class PackageSlot(entities.DlEntity):
    """
    Webhook object
    """
    # platform
    display_name: str = entities.DlProperty(location=['displayName'],
                                            _type=str)
    display_scopes: typing.Union[typing.List['entities.SlotDisplayScope'], None] = entities.DlProperty(
        location=['displayScopes'],
        _kls='SlotDisplayScope')
    module_name: str = entities.DlProperty(location=['moduleName'],
                                           _type=str,
                                           default='default_module')
    function_name: str = entities.DlProperty(location=['functionName'],
                                             _type=str,
                                             default='run')
    display_icon: str = entities.DlProperty(location=['displayIcon'],
                                            _type=str)
    post_action: SlotPostAction = entities.DlProperty(location=['postAction'],
                                                      _type=str,
                                                      _kls='SlotPostAction')
    default_inputs: typing.Union[typing.List['entities.FunctionIO'], None] = entities.DlProperty(
        location=['defaultInputs'],
        _kls='FunctionIO')
    input_options: list = entities.DlProperty(
        location=['inputOptions'],
        _kls='FunctionIO')

    @post_action.default
    def post_action_setter(self):
        return SlotPostAction(type=SlotPostActionType.NO_ACTION)

    @classmethod
    def from_json(cls, _json):
        inst = cls(_dict=_json)
        return inst

    def to_json(self):
        _json = self._dict.copy()
        return _json



import logging
import attr
from enum import Enum

from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class HttpMethod(str, Enum):
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PATCH = "PATCH"


@attr.s
class Webhook(entities.BaseEntity):
    """
    Webhook object
    """
    # platform
    id = attr.ib()
    url = attr.ib()
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    creator = attr.ib()
    name = attr.ib()

    # name change
    project_id = attr.ib()
    http_method = attr.ib()
    hook_url = attr.ib()

    # SDK
    _client_api = attr.ib(type=ApiClient, repr=False)
    _project = attr.ib()

    # repos
    _webhooks = attr.ib(default=None)

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @classmethod
    def from_json(cls, _json: dict, client_api: ApiClient, project=None):
        """
        :param _json: platform json
        :param client_api: ApiClient entity
        :param project: project entity
        :return:
        """
        if project is not None:
            if project.id != _json.get('project', None):
                logger.warning('Webhook has been fetched from a project that is not in it projects list')
                project = None
        return cls(
            http_method=_json.get('httpMethod', None),
            created_at=_json.get("createdAt", None),
            updated_at=_json.get("updatedAt", None),
            project_id=_json.get('project', None),
            hook_url=_json.get('hookUrl', None),
            creator=_json.get("creator", None),
            name=_json.get("name", None),
            url=_json.get("url", None),
            id=_json.get("id", None),
            client_api=client_api,
            project=project
        )

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def webhooks(self):
        if self._webhooks is None:
            self._webhooks = repositories.Webhooks(client_api=self._client_api, project=self._project)
        assert isinstance(self._webhooks, repositories.Webhooks)
        return self._webhooks

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(
            self,
            filter=attr.filters.exclude(
                attr.fields(Webhook)._project,
                attr.fields(Webhook)._client_api,
                attr.fields(Webhook).project_id,
                attr.fields(Webhook).hook_url,
                attr.fields(Webhook).http_method,
                attr.fields(Webhook).created_at,
                attr.fields(Webhook).updated_at,
            ),
        )

        _json['project'] = self.project_id
        _json['hookUrl'] = self.hook_url
        _json['httpMethod'] = self.http_method
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at

        return _json

    def delete(self):
        return self.webhooks.delete(self)

    def update(self):
        return self.webhooks.update(self)



def main():
    import dtlpy as dl

    dataset = dl.projects.get("project_name").datasets.get("dataset_name")

    ############################
    # convert annotations list #
    ############################

    # convert from dataloop to other formats #
    ##########################################

    # known format #

    item = dataset.items.get(item_id="item_id")
    annotations = item.annotations.list()
    converter = dl.Converter()
    converted_annotations = converter.convert(annotations=annotations,
                                              from_format='dataloop',
                                              to_format='yolo')

    # custom format #

    # custom conversion function
    # converts 1 dataloop annotation to custom format annotation
    # returns converted annotation
    def my_converter(annotation):
        """
        :param annotation: dataloop Annotation object
        :return: format of new annotation
        """
        ann = {'label': annotation.label, 'type': annotation.type}
        return ann

    converted_annotations = converter.convert(annotations=annotations,
                                              from_format='dataloop',
                                              to_format='my_format',
                                              conversion_func=my_converter)

    # convert from other formats to dataloop format #
    #################################################

    # known format #

    # load yolo annotations from file
    annotations = list()
    with open('annotations_filepath.txt', 'r') as f:
        line = f.readline()
        while line:
            annotations.append(line)
            line = f.readline()

    # create converter object
    converter = dl.Converter()

    # convert
    converted_annotations = converter.convert(annotations=annotations,
                                              from_format='yolo',
                                              to_format='dataloop')

    # custom format #

    # load yolo annotations from file
    annotations = list()
    with open('annotations_filepath.txt', 'r') as f:
        line = f.readline()
        while line:
            annotations.append(line)
            line = f.readline()

    # create converter object
    converter = dl.Converter()

    # custom conversion function
    # converts 1 custom annotation to dataloop annotation
    # dataloop annotation
    def my_converter(annotation):
        """
        :param annotation: custom annotation format
        :type annotation: dict
        :return: dataloop Annotation object
        """
        annotations = dl.Annotation.new(annotation_definition=dl.Box(top=annotation['top'],
                                                                     bottom=annotation['bottom'],
                                                                     left=annotation['left'],
                                                                     right=annotation['right'],
                                                                     label=annotation['label']))

        return annotation

    converted_annotations = converter.convert(annotations=annotations,
                                              from_format='my_format',
                                              to_format='dataloop',
                                              conversion_func=my_converter)

    ###############################
    # convert dataset annotations #
    ###############################

    # known format #
    ################

    converter = dl.Converter()

    converter.save_to_format = 'json'
    converter.convert_dataset(dataset=dataset,
                              to_format='coco',
                              local_path='some/local/path/to/download/converted/annotations')

    # custom format #
    #################

    def my_converter(annotation):
        """
        :param annotation: dataloop Annotation object
        :return: format of new annotation
        """
        ann = {'label': annotation.label, 'type': annotation.type}
        return ann

    converter = dl.Converter()

    converter.save_to_format = 'json'
    converter.convert_dataset(dataset=dataset,
                              to_format='my_format',
                              conversion_func=my_converter,
                              local_path='some/local/path/to/download/converted/annotations')

    ################
    # save to file #
    ################

    item = dataset.items.get(item_id="item_id")
    annotations = item.annotations.list()
    converter = dl.Converter()
    converter.convert(annotations=annotations,
                      from_format='dataloop',
                      to_format='yolo')

    # what file format to save to
    converter.save_to_format = '.txt'
    # save
    converter.save_to_file(save_to='local_path', to_format='yolo')



"""
Dataloop cookie state
"""

import os
import time
import json
import logging
import random
from .service_defaults import DATALOOP_PATH
from filelock import FileLock

logger = logging.getLogger(name='dtlpy')

NUM_TRIES = 3


class CookieIO:
    """
    Cookie interface for Dataloop parameters
    """

    def __init__(self, path, create=True, local=False):
        self.COOKIE = path
        self.local = local
        if create:
            self.create()

    @staticmethod
    def init():
        global_cookie_file = os.path.join(DATALOOP_PATH, 'cookie.json')
        return CookieIO(global_cookie_file)

    @staticmethod
    def init_local_cookie(create=False):
        local_cookie_file = os.path.join(os.getcwd(), '.dataloop', 'state.json')
        return CookieIO(local_cookie_file, create=create, local=True)

    @staticmethod
    def init_package_json_cookie(create=False):
        package_json_file = os.path.join(os.getcwd(), 'package.json')
        return CookieIO(package_json_file, create=create, local=True)

    def create(self):
        # create directory '.dataloop' if not exists
        if not os.path.isdir(os.path.dirname(self.COOKIE)):
            os.makedirs(os.path.dirname(self.COOKIE))

        if not os.path.isfile(self.COOKIE) or os.path.getsize(self.COOKIE) == 0:
            logger.debug('COOKIE.create: File: {}'.format(self.COOKIE))
            self.reset()
        try:
            with FileLock(self.COOKIE + ".lock"):
                with open(self.COOKIE, 'r') as f:
                    json.load(f)
        except ValueError:
            print('FATAL ERROR: COOKIE {!r} is corrupted. please fix or delete the file.'.format(self.COOKIE))
            raise SystemExit

    def read_json(self, create=False):
        # which cookie
        if self.local:
            self.COOKIE = os.path.join(os.getcwd(), '.dataloop', 'state.json')

        # check if file exists - and create
        if not os.path.isfile(self.COOKIE) and create:
            self.create()

        # check if file exists
        if not os.path.isfile(self.COOKIE):
            logger.debug('COOKIE.read: File does not exist: {}. Return None'.format(self.COOKIE))
            cfg = {}
        else:
            # read cookie
            cfg = {}
            for i in range(NUM_TRIES):
                try:
                    with FileLock(self.COOKIE + ".lock"):
                        with open(self.COOKIE, 'r') as fp:
                            cfg = json.load(fp)
                    break
                except Exception:
                    if i == (NUM_TRIES - 1):
                        raise
                    time.sleep(random.random())
                    continue
        return cfg

    def get(self, key):
        if key not in ['calls_counter']:
            # ignore logging for some keys
            logger.debug('COOKIE.read: key: {}'.format(key))
        cfg = self.read_json()
        if key in cfg.keys():
            value = cfg[key]
        else:
            logger.debug(msg='Key not in platform cookie file: {}. Return None'.format(key))
            value = None
        return value

    def put(self, key, value):
        if key not in ['calls_counter']:
            # ignore logging for some keys
            logger.debug('COOKIE.write: key: {}'.format(key))
        # read and write
        cfg = self.read_json(create=True)
        cfg[key] = value
        with FileLock(self.COOKIE + ".lock"):
            with open(self.COOKIE, 'w') as fp:
                json.dump(cfg, fp, indent=2)

    def reset(self):
        with FileLock(self.COOKIE + ".lock"):
            with open(self.COOKIE, 'w') as fp:
                json.dump({}, fp, indent=2)



import logging
import typing
import json
from enum import Enum
from .. import exceptions, entities

logger = logging.getLogger(name='dtlpy')


class PackageInputType(str, Enum):
    DATASET = "Dataset"
    ITEM = "Item"
    ANNOTATION = "Annotation"
    EXECUTION = "Execution"
    TASK = "Task"
    ASSIGNMENT = "Assignment"
    SERVICE = "Service"
    PACKAGE = "Package"
    PROJECT = "Project"
    RECIPE = "Recipe"
    JSON = "Json"
    STRING = "String"
    NUMBER = "Number"
    INT = "Integer"
    FLOAT = "Float"
    BOOLEAN = "Boolean"
    MODEL = "Model"
    DATASETS = "Dataset[]"
    ITEMS = "Item[]"
    ANNOTATIONS = "Annotation[]"
    EXECUTIONS = "Execution[]"
    TASKS = "Task[]"
    ASSIGNMENTS = "Assignment[]"
    SERVICES = "Service[]"
    PACKAGES = "Package[]"
    PROJECTS = "Project[]"
    JSONS = "Json[]"
    STRINGS = "String[]"
    NUMBERS = "Number[]"
    INTS = "Integer[]"
    FLOATS = "Float[]"
    BOOLEANS = "Boolean[]"
    MODELS = "Model[]"
    RECIPES = "Recipe[]"


class PackageFunction(entities.DlEntity):
    """
    Webhook object
    """

    # platform
    name: str = entities.DlProperty(location=['name'],
                                    _type=str,
                                    default=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME)
    description: str = entities.DlProperty(location=['description'],
                                           _type=str)
    display_name: typing.Union[str, None] = entities.DlProperty(location=['displayName'],
                                                                _type=typing.Union[str, None])
    display_icon: typing.Union[str, None] = entities.DlProperty(location=['displayIcon'],
                                                                _type=typing.Union[str, None])

    outputs: typing.Union[typing.List['entities.FunctionIO'], None] = entities.DlProperty(location=['output'],
                                                                                          _kls='FunctionIO')
    inputs: typing.Union[typing.List['entities.FunctionIO'], None] = entities.DlProperty(location=['input'],
                                                                                         _kls='FunctionIO')
    compute_config: str = entities.DlProperty(location=['computeConfig'], _type=str, default=None)

    def __repr__(self):
        # TODO need to move to DlEntity
        return f"PackageFunction(name={self.name}, description={self.description})"

    @classmethod
    def from_json(cls, _json):
        inst = cls(_dict=_json)
        return inst

    @outputs.default
    def get_outputs(self):
        outputs = list()
        return outputs

    @inputs.default
    def get_inputs(self):
        inputs = list()
        return inputs

    def to_json(self):
        _json = self._dict.copy()
        return _json


class FunctionIO(entities.DlEntity):
    INPUT_TYPES: list = [val for key, val in PackageInputType.__dict__.items() if not key.startswith('_')]
    type = entities.DlProperty(location=['type'], _type=str)
    value = entities.DlProperty(location=['value'], _type=str)
    name = entities.DlProperty(location=['name'], _type=str)
    actions = entities.DlProperty(location=['actions'], _type=list)
    description = entities.DlProperty(location=['description'], _type=str)
    integration = entities.DlProperty(location=['integration'], _type=dict)
    mandatory = entities.DlProperty(location=['mandatory'], _type=bool)

    def __repr__(self):
        # TODO need to move to DlEntity
        return f"FunctionIO(type={self.type}, name={self.name}, value={self.value})"

    @name.default
    def set_name(self):
        if self.type == PackageInputType.ITEM:
            return 'item'
        elif self.type == PackageInputType.DATASET:
            return 'dataset'
        elif self.type == PackageInputType.ANNOTATION:
            return 'annotation'
        elif self.type == PackageInputType.PROJECT:
            return 'project'
        elif self.type == PackageInputType.PACKAGE:
            return 'package'
        elif self.type == PackageInputType.SERVICE:
            return 'service'
        elif self.type == PackageInputType.EXECUTION:
            return 'execution'
        elif self.type == PackageInputType.MODEL:
            return 'model'
        else:
            return 'config'

    @type.validator
    def check_type(self, value):
        if value not in self.INPUT_TYPES:
            raise exceptions.PlatformException(
                error='400',
                message='Invalid input type: {!r}. Please one from dl.PackageInputType'.format(value))

    @staticmethod
    def is_json_serializable(val):
        try:
            json.dumps(val)
            is_json_serializable = True
        except Exception:
            is_json_serializable = False
        return is_json_serializable

    @value.validator
    def check_value(self, value):
        value_ok = True
        expected_value = 'Expected value should be:'
        if self.type == PackageInputType.JSON:
            expected_value = '{} json serializable'.format(expected_value)
            if not self.is_json_serializable(value):
                value_ok = False
        else:
            expected_value = 'Unknown value type: {}'.format(type(value))
            if type(value) not in [dict, str, float, int, bool, list]:
                value_ok = False

        if not value_ok and value is not None:
            raise exceptions.PlatformException('400', 'Illegal value. {}'.format(expected_value))

    def to_json(self, resource='package'):
        """
        :param resource:
        """
        if resource == 'package':
            _json = {
                'name': self.name,
                'type': self.type,
            }
            if self.actions:
                _json['actions'] = self.actions
            if self.description:
                _json['description'] = self.description
        elif resource in ['execution', 'service']:
            _json = {
                self.name: self.value
            }
        else:
            raise exceptions.PlatformException('400', 'Please select resource from: package, execution')

        return _json

    @classmethod
    def from_json(cls, _json):
        inst = cls(_dict=_json)
        return inst



def main(first_project_name, second_project_name, first_dataset_name, second_dataset_name, first_remote_filepath, second_remote_filepath):
    """
    Copy annotations between items
    :return:
    """
    import dtlpy as dl

    # FROM get the annotations from item
    project = dl.projects.get(project_name=first_project_name)
    dataset = project.datasets.get(dataset_name=first_dataset_name)
    item = dataset.items.get(filepath=first_remote_filepath)

    # get annotations
    annotations = item.annotations.list()

    # TO post annotations to other item
    project = dl.projects.get(project_name=second_project_name)
    dataset = project.datasets.get(dataset_name=second_dataset_name)
    item = dataset.items.get(filepath=second_remote_filepath)

    # post
    item.annotations.upload(annotations=annotations)



import sys
from collections import deque
import validators
import traceback
import tempfile
import requests
import asyncio
import logging
import pandas
import shutil
import json
import time
import tqdm
import os
import io
import numpy as np
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
from PIL import Image

from . import upload_element

from .. import PlatformException, entities, repositories, exceptions
from ..services import Reporter

logger = logging.getLogger(name='dtlpy')

NUM_TRIES = 5  # try to upload 3 time before fail on item


class Uploader:
    def __init__(self, items_repository: repositories.Items, output_entity=entities.Item, no_output=False):
        assert isinstance(items_repository, repositories.Items)
        self.items_repository = items_repository
        self.remote_url = "/datasets/{}/items".format(self.items_repository.dataset.id)
        self.__stop_create_existence_dict = False
        self.mode = 'skip'
        self.num_files = 0
        self.i_item = 0
        self.pbar = tqdm.tqdm(total=0,
                              disable=self.items_repository._client_api.verbose.disable_progress_bar_upload_items,
                              file=sys.stdout, desc='Upload Items')
        self.reporter = Reporter(num_workers=0,
                                 resource=Reporter.ITEMS_UPLOAD,
                                 print_error_logs=items_repository._client_api.verbose.print_error_logs,
                                 output_entity=output_entity,
                                 client_api=items_repository._client_api,
                                 no_output=no_output)

    def upload(
            self,
            # what to upload
            local_path,
            local_annotations_path=None,
            # upload options
            remote_path=None,
            remote_name=None,
            file_types=None,
            overwrite=False,
            item_metadata=None,
            export_version: str = entities.ExportVersion.V1,
            item_description=None,
            raise_on_error=False,
            return_as_list=False
    ):
        """
        Upload local file to dataset.
        Local filesystem will remain.
        If `*` at the end of local_path (e.g. '/images/*') items will be uploaded without head directory

        :param local_path: local file or folder to upload
        :param local_annotations_path: path to Dataloop format annotations json files.
        :param remote_path: remote path to save.
        :param remote_name: remote base name to save.
        :param file_types: list of file type to upload. e.g ['.jpg', '.png']. default is all
        :param overwrite: optional - default = False
        :param item_metadata: upload the items with the metadata dictionary
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :param str item_description: add a string description to the uploaded item
        :param bool raise_on_error: raise an exception if an error occurs
        :param bool return_as_list: always return a list of items 

        :return: Output (list)
        """
        ###################
        # Default options #
        ###################
        if overwrite:
            self.mode = 'overwrite'
        if isinstance(local_path, pandas.DataFrame):
            futures = self._build_elements_from_df(local_path)
        else:
            start_time = time.time()
            logger.debug(f"Building elements from inputs started: start time: {start_time}")
            futures = self._build_elements_from_inputs(local_path=local_path,
                                                       local_annotations_path=local_annotations_path,
                                                       # upload options
                                                       remote_path=remote_path,
                                                       remote_name=remote_name,
                                                       file_types=file_types,
                                                       item_metadata=item_metadata,
                                                       export_version=export_version,
                                                       item_description=item_description)
            logger.debug(f"Building elements from inputs completed:  time taken: {time.time() - start_time}")
        num_files = len(futures)
        while futures:
            futures.popleft().result()
        logger.info("Uploading {} items..".format(num_files))
        self.pbar.close()
        # summary
        logger.info("Number of total files: {}".format(num_files))
        status_list = self.reporter.status_list
        for action in set(status_list):
            n_for_action = self.reporter.status_count(status=action)
            logger.info("Number of files {}: {}".format(action, n_for_action))

        # log error
        errors_count = self.reporter.failure_count
        if errors_count > 0:
            error_text = ""
            log_filepath = self.reporter.generate_log_files()
            # Get up to 5 error examples for the exception message
            if self.reporter._errors:
                error_examples = list(self.reporter._errors.values())[:5]
                error_text = " | ".join(error_examples)
            error_message = f"Errors in {errors_count} files. Errors: {error_text}" 
            if log_filepath is not None:
                error_message += f", see {log_filepath} for full log"
            if raise_on_error is True:
                raise PlatformException(
                    error="400", message=error_message
                )
            else:
                logger.warning(error_message)

        if return_as_list is True:
            # return list of items
            return list(self.reporter.output)
        if len(status_list) == 1:
            # if there is only one item, return it
            try:
                return next(self.reporter.output)
            except StopIteration:
                # if there is no items, return None
                return None
        # if there are multiple items, return the generator
        return self.reporter.output

    def _build_elements_from_inputs(self,
                                    local_path,
                                    local_annotations_path,
                                    # upload options
                                    remote_path,
                                    file_types,
                                    remote_name,
                                    item_metadata,
                                    export_version: str = entities.ExportVersion.V1,
                                    item_description=None):
        # fix remote path
        if remote_path is None:
            if isinstance(local_path, str) and local_path.startswith('external://'):
                remote_path = None
            else:
                remote_path = "/"
        if remote_path and not remote_path.startswith('/'):
            remote_path = f"/{remote_path}"
        if remote_path and not remote_path.endswith("/"):
            remote_path = f"{remote_path}/"

        if remote_name:
            remote_name = remote_name.lstrip('/')

        if file_types is not None and not isinstance(file_types, list):
            msg = '"file_types" should be a list of file extension. e.g [".jpg", ".png"]'
            raise PlatformException(error="400", message=msg)
        if item_metadata is not None:
            if not isinstance(item_metadata, dict) and not isinstance(item_metadata, entities.ExportMetadata):
                msg = '"item_metadata" should be a metadata dictionary. Got type: {}'.format(type(item_metadata))
                raise PlatformException(error="400", message=msg)
        if item_description is not None:
            if not isinstance(item_description, str):
                msg = '"item_description" should be a string. Got type: {}'.format(type(item_description))
                raise PlatformException(error="400", message=msg)

        ##########################
        # Convert inputs to list #
        ##########################
        local_annotations_path_list = None
        remote_name_list = None
        if not isinstance(local_path, list):
            local_path_list = [local_path]
            if remote_name is not None:
                if not isinstance(remote_name, str):
                    raise PlatformException(error="400",
                                            message='remote_name must be a string, got: {}'.format(type(remote_name)))
                remote_name_list = [remote_name]
            if local_annotations_path is not None:
                if not isinstance(local_annotations_path, str):
                    raise PlatformException(error="400",
                                            message='local_annotations_path must be a string, got: {}'.format(
                                                type(local_annotations_path)))
                local_annotations_path_list = [local_annotations_path]
        else:
            local_path_list = local_path
            if remote_name is not None:
                if not isinstance(remote_name, list):
                    raise PlatformException(error="400",
                                            message='remote_name must be a list, got: {}'.format(type(remote_name)))
                if not len(remote_name) == len(local_path_list):
                    raise PlatformException(error="400",
                                            message='remote_name and local_path_list must be of same length. '
                                                    'Received: remote_name: {}, '
                                                    'local_path_list: {}'.format(len(remote_name),
                                                                                 len(local_path_list)))
                remote_name_list = remote_name
            if local_annotations_path is not None:
                if not len(local_annotations_path) == len(local_path_list):
                    raise PlatformException(error="400",
                                            message='local_annotations_path and local_path_list must be of same lenght.'
                                                    ' Received: local_annotations_path: {}, '
                                                    'local_path_list: {}'.format(len(local_annotations_path),
                                                                                 len(local_path_list)))
                local_annotations_path_list = local_annotations_path

        if local_annotations_path is None:
            local_annotations_path_list = [None] * len(local_path_list)

        if remote_name is None:
            remote_name_list = [None] * len(local_path_list)

        futures = deque()
        total_size = 0
        for upload_item_element, remote_name, upload_annotations_element in zip(local_path_list,
                                                                                remote_name_list,
                                                                                local_annotations_path_list):
            if isinstance(upload_item_element, np.ndarray):
                # convert numpy.ndarray to io.BytesI
                if remote_name is None:
                    raise PlatformException(
                        error="400",
                        message='Upload element type was numpy.ndarray. providing param "remote_name" is mandatory')
                file_extension = os.path.splitext(remote_name)
                if file_extension[1].lower() in ['.jpg', '.jpeg']:
                    item_format = 'JPEG'
                elif file_extension[1].lower() == '.png':
                    item_format = 'PNG'
                else:
                    raise PlatformException(
                        error="400",
                        message='"remote_name" with  .jpg/.jpeg or .png extension are supported '
                                'when upload element of numpy.ndarray type.')

                buffer = io.BytesIO()
                Image.fromarray(upload_item_element).save(buffer, format=item_format)
                buffer.seek(0)
                buffer.name = remote_name
                upload_item_element = buffer

            all_upload_elements = {
                'upload_item_element': upload_item_element,
                'total_size': total_size,
                'remote_name': remote_name,
                'remote_path': remote_path,
                'upload_annotations_element': upload_annotations_element,
                'item_metadata': item_metadata,
                'annotations_filepath': None,
                'with_head_folder': None,
                'filename': None,
                'root': None,
                'export_version': export_version,
                'item_description': item_description,
                'driver_path': None
            }
            if isinstance(upload_item_element, str):
                with_head_folder = True
                if upload_item_element.endswith('*'):
                    with_head_folder = False
                    upload_item_element = os.path.dirname(upload_item_element)
                    all_upload_elements['upload_item_element'] = upload_item_element

                if os.path.isdir(upload_item_element):
                    for root, subdirs, files in os.walk(upload_item_element):
                        for filename in files:
                            all_upload_elements['with_head_folder'] = with_head_folder
                            all_upload_elements['filename'] = filename
                            all_upload_elements['root'] = root
                            _, ext = os.path.splitext(filename)
                            if file_types is None or ext in file_types:
                                upload_elem = upload_element.DirUploadElement(all_upload_elements=all_upload_elements)
                                futures.append(self.upload_single_element(upload_elem))
                    continue

                # add single file
                elif os.path.isfile(upload_item_element):
                    upload_elem = upload_element.FileUploadElement(all_upload_elements=all_upload_elements)

                elif upload_item_element.startswith('external://'):
                    try:
                        driver_path = repositories.Drivers.get(driver_id=self.items_repository.dataset.driver).path
                        all_upload_elements['driver_path'] = driver_path
                    except Exception:
                        logger.error("Attempting to upload external item without driver path. This may cause issues.")
                    upload_elem = upload_element.ExternalItemUploadElement(all_upload_elements=all_upload_elements)

                elif self.is_url(upload_item_element):
                    upload_elem = upload_element.UrlUploadElement(all_upload_elements=all_upload_elements)

                else:
                    raise PlatformException("404", "Unknown local path: {}".format(local_path))

            elif isinstance(upload_item_element, entities.Item):
                upload_elem = upload_element.ItemLinkUploadElement(all_upload_elements=all_upload_elements)

            elif isinstance(upload_item_element, entities.Link):
                upload_elem = upload_element.LinkUploadElement(all_upload_elements=all_upload_elements)

            elif isinstance(upload_item_element, entities.PromptItem):
                upload_elem = upload_element.PromptUploadElement(all_upload_elements=all_upload_elements)

            elif isinstance(upload_item_element, entities.ItemGis):
                buffer = io.BytesIO(json.dumps(upload_item_element.to_json()).encode('utf-8'))
                buffer.name = upload_item_element.name
                all_upload_elements['upload_item_element'] = buffer
                upload_elem = upload_element.BinaryUploadElement(all_upload_elements=all_upload_elements)

            elif isinstance(upload_item_element, bytes) or \
                    isinstance(upload_item_element, io.BytesIO) or \
                    isinstance(upload_item_element, io.BufferedReader) or \
                    isinstance(upload_item_element, io.TextIOWrapper):
                upload_elem = upload_element.BinaryUploadElement(all_upload_elements=all_upload_elements)
                # get size from binaries
                try:
                    total_size += upload_item_element.__sizeof__()
                except Exception:
                    logger.warning("Cant get binaries size")

            else:
                raise PlatformException(
                    error="400",
                    message=f"Unknown element type to upload ('local_path'). received type: {type(upload_item_element)}. "
                            "known types (or list of those types): str (dir, file, url), bytes, io.BytesIO, "
                            "numpy.ndarray, io.TextIOWrapper, Dataloop.Item, Dataloop.Link")

            futures.append(self.upload_single_element(upload_elem))
        return futures

    def upload_single_element(self, elem):
        """
        upload a signal element
        :param elem: UploadElement
        """
        self.num_files += 1
        self.i_item += 1
        self.pbar.total += 1
        self.reporter.upcount_num_workers()
        future = asyncio.run_coroutine_threadsafe(
            self.__upload_single_item_wrapper(element=elem,
                                              mode=self.mode,
                                              pbar=self.pbar,
                                              reporter=self.reporter),
            loop=self.items_repository._client_api.event_loop.loop)
        return future

    def _build_elements_from_df(self, df: pandas.DataFrame):
        futures = deque()
        for index, row in df.iterrows():
            # DEFAULTS
            elem = {'local_annotations_path': None,
                    'remote_path': None,
                    'remote_name': None,
                    'file_types': None,
                    'item_metadata': None,
                    'item_description': None}
            elem.update(row)
            future = self._build_elements_from_inputs(**elem)
            # append deque using +
            futures += future
        return futures

    async def __single_external_sync(self, element):
        storage_id = element.buffer.split('//')[1]
        req_json = dict()
        req_json['filename'] = element.remote_filepath
        req_json['storageId'] = storage_id
        success, response = self.items_repository._client_api.gen_request(req_type='post',
                                                                          path='/datasets/{}/imports'.format(
                                                                              self.items_repository.dataset.id),
                                                                          json_req=[req_json])

        if success:
            items = entities.Item.from_json(client_api=self.items_repository._client_api, _json=response.json()[0],
                                            project=self.items_repository._dataset._project,
                                            dataset=self.items_repository.dataset)
        else:
            raise exceptions.PlatformException(response)
        return items, response.headers.get('x-item-op', 'na')

    async def __single_async_upload(self,
                                    filepath,
                                    remote_path,
                                    uploaded_filename,
                                    last_try,
                                    mode,
                                    item_metadata,
                                    callback,
                                    item_description
                                    ):
        """
        Upload an item to dataset

        :param filepath: local filepath of the item
        :param remote_path: remote directory of filepath to upload
        :param uploaded_filename: optional - remote filename
        :param last_try: print log error only if last try
        :param mode: 'skip'  'overwrite'
        :param item_metadata: item metadata
        :param str item_description: add a string description to the uploaded item
        :param callback:
        :return: Item object
        """

        need_close = False
        if isinstance(filepath, str):
            # upload local file
            if not os.path.isfile(filepath):
                raise PlatformException(error="404", message="Filepath doesnt exists. file: {}".format(filepath))
            if uploaded_filename is None:
                uploaded_filename = os.path.basename(filepath)
            if os.path.isfile(filepath):
                item_type = 'file'
            else:
                item_type = 'dir'
            item_size = os.stat(filepath).st_size
            to_upload = open(filepath, 'rb')
            need_close = True

        else:
            # upload from buffer
            if isinstance(filepath, bytes):
                to_upload = io.BytesIO(filepath)
            elif isinstance(filepath, io.BytesIO):
                to_upload = filepath
            elif isinstance(filepath, io.BufferedReader):
                to_upload = filepath
            elif isinstance(filepath, io.TextIOWrapper):
                to_upload = filepath
            else:
                raise PlatformException("400", "Unknown input filepath type received: {}".format(type(filepath)))

            if uploaded_filename is None:
                if hasattr(filepath, "name"):
                    uploaded_filename = filepath.name
                else:
                    raise PlatformException(error="400",
                                            message="Must have filename when uploading bytes array (uploaded_filename)")

            item_size = to_upload.seek(0, 2)
            to_upload.seek(0)
            item_type = 'file'
        try:
            response = await self.items_repository._client_api.upload_file_async(to_upload=to_upload,
                                                                                 item_type=item_type,
                                                                                 item_size=item_size,
                                                                                 item_metadata=item_metadata,
                                                                                 remote_url=self.remote_url,
                                                                                 uploaded_filename=uploaded_filename,
                                                                                 remote_path=remote_path,
                                                                                 callback=callback,
                                                                                 mode=mode,
                                                                                 item_description=item_description)
        except Exception:
            raise
        finally:
            if need_close:
                to_upload.close()

        if response.ok:
            if item_size != response.json().get('metadata', {}).get('system', {}).get('size', 0):
                self.items_repository.delete(item_id=response.json()['id'])
                raise PlatformException(500,
                                        "The uploaded file is corrupted. "
                                        "Please try again. If it happens again please contact support.")
            item = self.items_repository.items_entity.from_json(client_api=self.items_repository._client_api,
                                                                _json=response.json(),
                                                                dataset=self.items_repository.dataset)
        else:
            raise PlatformException(response)
        return item, response.headers.get('x-item-op', 'na')

    async def __upload_single_item_wrapper(self, element, pbar, reporter, mode):
        async with self.items_repository._client_api.event_loop.semaphore('items.upload', 5):
            # assert isinstance(element, UploadElement)
            item = False
            err = None
            trace = None
            saved_locally = False
            temp_dir = None
            action = 'na'
            remote_folder, remote_name = os.path.split(element.remote_filepath)

            if element.type == 'url':
                saved_locally, element.buffer, temp_dir = self.url_to_data(element.buffer)
            elif element.type == 'link':
                element.buffer = self.link(ref=element.buffer.ref, dataset_id=element.buffer.dataset_id,
                                           type=element.buffer.type, mimetype=element.buffer.mimetype)

            for i_try in range(NUM_TRIES):
                try:
                    logger.debug("Upload item: {path}. Try {i}/{n}. Starting..".format(path=remote_name,
                                                                                       i=i_try + 1,
                                                                                       n=NUM_TRIES))
                    if element.type == 'external_file':
                        item, action = await self.__single_external_sync(element)
                    else:
                        if element.annotations_filepath is not None and \
                                element.item_metadata == entities.ExportMetadata.FROM_JSON:
                            element.item_metadata = {}
                            with open(element.annotations_filepath) as ann_f:
                                item_metadata = json.load(ann_f)
                            if 'metadata' in item_metadata:
                                element.item_metadata = item_metadata['metadata']
                        item, action = await self.__single_async_upload(filepath=element.buffer,
                                                                        mode=mode,
                                                                        item_metadata=element.item_metadata,
                                                                        remote_path=remote_folder,
                                                                        uploaded_filename=remote_name,
                                                                        last_try=(i_try + 1) == NUM_TRIES,
                                                                        callback=None,
                                                                        item_description=element.item_description)
                    logger.debug("Upload item: {path}. Try {i}/{n}. Success. Item id: {id}".format(path=remote_name,
                                                                                                   i=i_try + 1,
                                                                                                   n=NUM_TRIES,
                                                                                                   id=item.id))
                    if isinstance(item, entities.Item):
                        break
                    time.sleep(0.3 * (2 ** i_try))
                except Exception as e:
                    err = e
                    trace = traceback.format_exc()
                    logger.debug("Upload item: {path}. Try {i}/{n}. Fail.\n{trace}".format(path=remote_name,
                                                                                           i=i_try + 1,
                                                                                           n=NUM_TRIES,
                                                                                           trace=trace))

                finally:
                    if saved_locally and os.path.isdir(temp_dir):
                        shutil.rmtree(temp_dir)
            if item:
                if action in ['overwrite', 'created'] and element.annotations_filepath is not None:
                    try:
                        await self.__async_upload_annotations(annotations_filepath=element.annotations_filepath,
                                                              item=item)
                    except Exception:
                        logger.exception('Error uploading annotations to item id: {}'.format(item.id))

                reporter.set_index(status=action,
                                   output=item.to_json(),
                                   success=True,
                                   ref=item.id)
                if pbar is not None:
                    pbar.update()
                    self.items_repository._client_api.callbacks.run_on_event(
                        event=self.items_repository._client_api.callbacks.CallbackEvent.ITEMS_UPLOAD,
                        context={'item_id': item.id, 'dataset_id': item.dataset_id},
                        progress=round(pbar.n / pbar.total * 100, 0))
            else:
                if isinstance(element.buffer, str):
                    ref = element.buffer
                elif hasattr(element.buffer, "name"):
                    ref = element.buffer.name
                else:
                    ref = 'Unknown'
                reporter.set_index(ref=ref, status='error',
                                   success=False,
                                   error="{}\n{}".format(err, trace))

    async def __async_upload_annotations(self, annotations_filepath, item):
        with open(annotations_filepath, 'r', encoding="utf8") as f:
            annotations = json.load(f)
        # wait for coroutines on the current event loop
        return await item.annotations._async_upload_annotations(annotations=annotations['annotations'])

    @staticmethod
    def url_to_data(url):
        chunk_size = 8192
        max_size = 30000000
        temp_dir = None

        # This will download the binaries from the URL user provided
        prepared_request = requests.Request(method='GET', url=url).prepare()
        with requests.Session() as s:
            retry = Retry(
                total=3,
                read=3,
                connect=3,
                backoff_factor=1,
            )
            adapter = HTTPAdapter(max_retries=retry)
            s.mount('http://', adapter)
            s.mount('https://', adapter)
            response = s.send(request=prepared_request, stream=True)

        total_length = response.headers.get("content-length")
        save_locally = int(total_length) > max_size

        if save_locally:
            # save to file
            temp_dir = tempfile.mkdtemp()
            temp_path = os.path.join(temp_dir, url.split('/')[-1].split('?')[0])
            with open(temp_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
            # save to output variable
            data = temp_path
        else:
            # save as byte stream
            data = io.BytesIO()
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    data.write(chunk)
            # go back to the beginning of the stream
            data.seek(0)
            data.name = url.split('/')[-1]

        return save_locally, data, temp_dir

    @staticmethod
    def is_url(url):
        try:
            return validators.url(url)
        except Exception:
            return False

    @staticmethod
    def link(ref, type, mimetype=None, dataset_id=None):
        """
        :param ref:
        :param type:
        :param mimetype:
        :param dataset_id:
        """
        link_info = {'type': type,
                     'ref': ref}

        if mimetype:
            link_info['mimetype'] = mimetype

        if dataset_id is not None:
            link_info['datasetId'] = dataset_id

        _json = {'type': 'link',
                 'shebang': 'dataloop',
                 'metadata': {'dltype': 'link',
                              'linkInfo': link_info}}

        uploaded_byte_io = io.BytesIO()
        uploaded_byte_io.write(json.dumps(_json).encode())
        uploaded_byte_io.seek(0)

        return uploaded_byte_io



import numpy as np
from . import BaseAnnotationDefinition
from .polygon import Polygon


class Box(BaseAnnotationDefinition):
    """
        Box annotation object
        Can create a box using 2 point using: "top", "left", "bottom", "right" (to form a box [(left, top), (right, bottom)])
        For rotated box add the "angel"
    """
    type = "box"

    def __init__(self,
                 left=None, top=None, right=None, bottom=None,
                 label=None, attributes=None, description=None, angle=None):
        """
        Can create a box using 2 point using:
         "top", "left", "bottom", "right" (to form a box [(left, top), (right, bottom)])
        And for create a rotated box need to add the angle of the rotation

        :param left: left x coordinate of the box
        :param top: top Y coordinate of the box
        :param right: right x coordinate of the box
        :param bottom: bottom Y coordinate of the box
        :param label: annotation label
        :param attributes: a list of attributes for the annotation
        :param description:
        :param angle: the angle of the rotation in degrees

        :return:
        """
        super().__init__(description=description, attributes=attributes)

        self.angle = angle
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom
        self.top_left = [left, top]
        self.top_right = [right, top]
        self.bottom_left = [left, bottom]
        self.bottom_right = [right, bottom]
        self.label = label

    @property
    def x(self):
        return [self.left, self.right]

    @property
    def y(self):
        return [self.top, self.bottom]

    @property
    def geo(self):
        return [
            [self.left, self.top],
            [self.right, self.bottom]
        ]

    @property
    def four_points(self):
        return [self.top_left, self.bottom_left, self.bottom_right, self.top_right]

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        if thickness is None:
            thickness = 2

        # draw annotation
        points = self.four_points

        # create image to draw on
        if alpha != 1:
            overlay = image.copy()
        else:
            overlay = image

        # draw annotation
        overlay = cv2.drawContours(
            image=overlay,
            contours=[np.round(points).astype(int)],
            contourIdx=-1,
            color=color,
            thickness=thickness,
            lineType=cv2.LINE_AA
        )

        if not isinstance(color, int) and len(color) == 4 and color[3] != 255:
            # add with opacity
            image = cv2.addWeighted(src1=overlay,
                                    alpha=alpha,
                                    src2=image,
                                    beta=1 - alpha,
                                    gamma=0)
        else:
            image = overlay

        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    def to_coordinates(self, color):
        points = [
            [self.left, self.top],
            [self.right, self.bottom]
        ]
        pts = [{"x": float(x), "y": float(y), "z": 0} for x, y in points]

        if self.angle is not None and self.angle != 0:
            pts.append(self.angle)

        return pts

    @staticmethod
    def from_coordinates(coordinates):
        return np.asarray([[pt["x"], pt["y"]] for pt in coordinates[:2]])

    @classmethod
    def from_json(cls, _json):
        coordinates = _json.get("coordinates", None) if "coordinates" in _json else _json.get("data", None)
        if coordinates is None:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        geo = cls.from_coordinates(coordinates=coordinates)

        left = np.min(geo[:, 0])
        top = np.min(geo[:, 1])
        right = np.max(geo[:, 0])
        bottom = np.max(geo[:, 1])

        angel = coordinates[2] if len(coordinates) > 2 and (
                isinstance(coordinates[2], float) or isinstance(coordinates[2], int)) else None

        return cls(
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            label=_json["label"],
            attributes=_json.get("attributes", None),
            angle=angel
        )

    @classmethod
    def from_segmentation(cls, mask, label, attributes=None):
        """
        Convert binary mask to Polygon

        :param mask: binary mask (0,1)
        :param label: annotation label
        :param attributes: annotations list of attributes

        :return: Box annotations list  to each separated  segmentation
        """
        polygons = Polygon.from_segmentation(
            mask=mask,
            label=label,
            attributes=attributes,
            max_instances=None
        )

        if not isinstance(polygons, list):
            polygons = [polygons]

        boxes = [
            cls(
                left=polygon.left,
                top=polygon.top,
                right=polygon.right,
                bottom=polygon.bottom,
                label=label,
                attributes=attributes
            ) for polygon in polygons
        ]

        return boxes



import traceback
from typing import Dict
from ..services.api_client import ApiClient
from .. import repositories
from .compute import ComputeContext, ComputeType

class ServiceDriver:
    def __init__(
            self,
            name: str,
            context: ComputeContext,
            compute_id: str,
            client_api: ApiClient,
            type: ComputeType = None,
            created_at: str = None,
            updated_at: str = None,
            namespace: str = None,
            metadata: Dict = None,
            url: str = None,
            archived: bool = None,
            id: str = None,
            is_cache_available: bool = None
    ):
        self.name = name
        self.context = context
        self.compute_id = compute_id
        self.client_api = client_api
        self.type = type or ComputeType.KUBERNETES
        self.created_at = created_at
        self.updated_at = updated_at
        self.namespace = namespace
        self.metadata = metadata
        self.url = url
        self.archived = archived
        self.id = id
        self.is_cache_available = is_cache_available
        self._service_drivers = None
        self._client_api = client_api

    @property
    def service_drivers(self):
        if self._service_drivers is None:
            self._service_drivers = repositories.ServiceDrivers(client_api=self._client_api)
        return self._service_drivers

    @staticmethod
    def _protected_from_json(_json: dict, client_api: ApiClient):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            service = ServiceDriver.from_json(_json=_json,
                                              client_api=client_api)
            status = True
        except Exception:
            service = traceback.format_exc()
            status = False
        return status, service

    @classmethod
    def from_json(cls, _json, client_api: ApiClient):
        return cls(
            name=_json.get('name'),
            context=ComputeContext.from_json(_json.get('context', dict())),
            compute_id=_json.get('computeId'),
            client_api=client_api,
            type=_json.get('type', None),
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            namespace=_json.get('namespace', None),
            metadata=_json.get('metadata', None),
            url=_json.get('url', None),
            archived=_json.get('archived', None),
            id=_json.get('id', None),
            is_cache_available=_json.get('isCacheAvailable', None)
        )

    def to_json(self):
        _json = {
            'name': self.name,
            'context': self.context.to_json(),
            'computeId': self.compute_id,
            'type': self.type,
        }
        if self.created_at is not None:
            _json['createdAt'] = self.created_at
        if self.updated_at is not None:
            _json['updatedAt'] = self.updated_at
        if self.namespace is not None:
            _json['namespace'] = self.namespace
        if self.metadata is not None:
            _json['metadata'] = self.metadata
        if self.url is not None:
            _json['url'] = self.url
        if self.archived is not None:
            _json['archived'] = self.archived
        if self.id is not None:
            _json['id'] = self.id
        if self.is_cache_available is not None:
            _json['isCacheAvailable'] = self.is_cache_available

        return _json

    def delete(self):
        """
        Delete a service driver
        """
        return self.service_drivers.delete(service_driver_id=self.id)

    def update(self):
        """
        Update a service driver
        """
        return self.service_drivers.update(service_driver=self)


import time
from typing import List
import logging
from urllib.parse import urlencode

from .. import entities, repositories, exceptions, miscellaneous
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')

MIN_INTERVAL = 1
BACKOFF_FACTOR = 1.2
MAX_INTERVAL = 12


class Models:
    """
    Models Repository
    """
    @staticmethod
    def _filter_to_dict(filter_obj):
        """Convert Filters object to dict, or return as-is if already dict/None"""
        if filter_obj is not None:
            filter_obj = filter_obj.prepare() if isinstance(filter_obj, entities.Filters) else filter_obj
        return filter_obj

    @staticmethod
    def _get_filter_from_model(model, subset_type, resource_type):
        """Extract filter dict from model metadata"""
        filter_dict = None
        if model is not None:
            if resource_type == entities.FiltersResource.ITEM:
                filter_dict = model.metadata.get('system', {}).get('subsets', {}).get(subset_type.value)
            else:  # ANNOTATION
                filter_dict = model.metadata.get('system', {}).get('annotationsSubsets', {}).get(subset_type.value)
        return filter_dict

    @staticmethod
    def _build_model_metadata(
        train_filter: entities.Filters = None,
        validation_filter: entities.Filters = None,
        annotations_train_filter: entities.Filters = None,
        annotations_validation_filter: entities.Filters = None,
        from_model: entities.Model = None
    ) -> dict:
        """
        Build model metadata with filters, optionally inheriting from existing model.
        
        :param train_filter: Training data filter (Filters object or dict)
        :param validation_filter: Validation data filter (Filters object or dict)
        :param annotations_train_filter: Training annotations filter (Filters object or dict)
        :param annotations_validation_filter: Validation annotations filter (Filters object or dict)
        :param from_model: Source model to inherit filters from (if not provided explicitly)
        :return: Metadata dictionary with filters
        """
        metadata = {'system': {'subsets': {}, 'annotationsSubsets': {}}}

        # Handle item filters
        train_filter_dict = Models._filter_to_dict(train_filter)
        if train_filter_dict is None and from_model is not None:
            train_filter_dict = Models._get_filter_from_model(
                model=from_model, subset_type=entities.DatasetSubsetType.TRAIN, resource_type=entities.FiltersResource.ITEM)

        validation_filter_dict = Models._filter_to_dict(validation_filter)
        if validation_filter_dict is None and from_model is not None:
            validation_filter_dict = Models._get_filter_from_model(
                model=from_model, subset_type=entities.DatasetSubsetType.VALIDATION, resource_type=entities.FiltersResource.ITEM)

        # Handle annotation filters
        annotations_train_filter_dict = Models._filter_to_dict(annotations_train_filter)
        if annotations_train_filter_dict is None and from_model is not None:
            annotations_train_filter_dict = Models._get_filter_from_model(
                model=from_model, subset_type=entities.DatasetSubsetType.TRAIN, resource_type=entities.FiltersResource.ANNOTATION)

        annotations_validation_filter_dict = Models._filter_to_dict(annotations_validation_filter)
        if annotations_validation_filter_dict is None and from_model is not None:
            annotations_validation_filter_dict = Models._get_filter_from_model(
                model=from_model, subset_type=entities.DatasetSubsetType.VALIDATION, resource_type=entities.FiltersResource.ANNOTATION)

        # Set filters in metadata
        if train_filter_dict is not None:
            metadata['system']['subsets']['train'] = train_filter_dict
        if validation_filter_dict is not None:
            metadata['system']['subsets']['validation'] = validation_filter_dict
        if annotations_train_filter_dict is not None:
            metadata['system']['annotationsSubsets']['train'] = annotations_train_filter_dict
        if annotations_validation_filter_dict is not None:
            metadata['system']['annotationsSubsets']['validation'] = annotations_validation_filter_dict

        return metadata

    def __init__(self,
                 client_api: ApiClient,
                 package: entities.Package = None,
                 project: entities.Project = None,
                 project_id: str = None):
        self._client_api = client_api
        self._project = project
        self._package = package
        self._project_id = project_id

        if self._project is not None:
            self._project_id = self._project.id

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._project_id is not None:
                projects = repositories.Projects(client_api=self._client_api)
                self._project = projects.get(project_id=self._project_id)
        if self._project is None:
            if self._package is not None:
                if self._package._project is not None:
                    self._project = self._package._project
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.models repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    @property
    def package(self) -> entities.Package:
        if self._package is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Package entity in {} repository.'.format(
                    self.__class__.__name__) +
                        ' Please use package.models or set a model')
        assert isinstance(self._package, entities.Package)
        return self._package

    ###########
    # methods #
    ###########
    def get(self, model_name=None, model_id=None) -> entities.Model:
        """
        Get model object
        :param model_name:
        :param model_id:
        :return: dl.Model object
        """

        if model_id is not None:
            success, response = self._client_api.gen_request(req_type="get",
                                                             path="/ml/models/{}".format(model_id))
            if not success:
                raise exceptions.PlatformException(response)
            model = entities.Model.from_json(client_api=self._client_api,
                                             _json=response.json(),
                                             project=self._project,
                                             package=self._package)
            # verify input model name is same as the given id
            if model_name is not None and model.name != model_name:
                logger.warning(
                    "Mismatch found in models.get: model_name is different then model.name:"
                    " {!r} != {!r}".format(
                        model_name,
                        model.name))
        elif model_name is not None:

            filters = entities.Filters(
                resource=entities.FiltersResource.MODEL,
                field='name',
                values=model_name
            )

            project_id = None

            if self._project is not None:
                project_id = self._project.id
            elif self._project_id is not None:
                project_id = self._project_id

            if project_id is not None:
                filters.add(field='projectId', values=project_id)

            if self._package is not None:
                filters.add(field='packageId', values=self._package.id)

            models = self.list(filters=filters)

            if models.items_count == 0:
                raise exceptions.PlatformException(
                    error='404',
                    message='Model not found. Name: {}'.format(model_name))
            elif models.items_count > 1:
                raise exceptions.PlatformException(
                    error='400',
                    message='More than one Model found by the name of: {}. Try "get" by id or "list()".'.format(
                        model_name))
            model = models.items[0]
        else:
            raise exceptions.PlatformException(
                error='400',
                message='No checked-out Model was found, must checkout or provide an identifier in inputs')

        return model

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Model]:
        jobs = [None for _ in range(len(response_items))]
        pool = self._client_api.thread_pools(pool_name='entity.create')

        # return triggers list
        for i_service, service in enumerate(response_items):
            jobs[i_service] = pool.submit(entities.Model._protected_from_json,
                                          **{'client_api': self._client_api,
                                             '_json': service,
                                             'package': self._package,
                                             'project': self._project})

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _list(self, filters: entities.Filters):
        # request
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path='/ml/models/query',
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List project model

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities
        """
        # default filters
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.MODEL)
        if self._project is not None:
            filters.add(field='projectId', values=self._project.id)
        if self._package is not None:
            filters.add(field='packageId', values=self._package.id)

        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))

        if filters.resource != entities.FiltersResource.MODEL:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.MODEL. Got: {!r}'.format(filters.resource))

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def _set_model_filter(self,
                          metadata: dict,
                          train_filter: entities.Filters = None,
                          validation_filter: entities.Filters = None):
        if metadata is None:
            metadata = {}
        if 'system' not in metadata:
            metadata['system'] = {}
        if 'subsets' not in metadata['system']:
            metadata['system']['subsets'] = {}
        if train_filter is not None:
            metadata['system']['subsets']['train'] = train_filter.prepare() if isinstance(train_filter,
                                                                                          entities.Filters) else train_filter
        if validation_filter is not None:
            metadata['system']['subsets']['validation'] = validation_filter.prepare() if isinstance(validation_filter,
                                                                                                    entities.Filters) else validation_filter
        return metadata

    @staticmethod
    def add_subset(
        model: entities.Model,
        subset_name: str,
        subset_filter=None,
        subset_annotation_filter=None,
    ):
        """
        Adds a subset for a model, specifying a subset of the model's dataset that could be used for training or
        validation. Optionally also adds an annotations subset.

        :param dtlpy.entities.Model model: the model to which the subset should be added
        :param str subset_name: the name of the subset
        :param subset_filter: filtering for items subset. Can be `entities.Filters`, `dict`, or `None`
        :param subset_annotation_filter: optional filtering for annotations subset. Can be `entities.Filters`, `dict`, or `None`

        Behavior:
        - If both filters are None, no metadata is added/changed.
        - If a filter is a dict, it is used as-is (no prepare()).
        - If a filter is `entities.Filters`, `.prepare()` is used.
        - Only non-None filters are added.
        """
        if subset_filter is None and subset_annotation_filter is None:
            return

        subset_filter_dict = subset_filter.prepare() if isinstance(subset_filter, entities.Filters) else subset_filter
        subset_annotation_filter_dict = (
            subset_annotation_filter.prepare()
            if isinstance(subset_annotation_filter, entities.Filters)
            else subset_annotation_filter
        )

        # Initialize containers only if needed
        if 'system' not in model.metadata:
            model.metadata['system'] = dict()
        if subset_filter_dict is not None:
            if 'subsets' not in model.metadata['system']:
                model.metadata['system']['subsets'] = dict()
            model.metadata['system']['subsets'][subset_name] = subset_filter_dict

        if subset_annotation_filter_dict is not None:
            if 'annotationsSubsets' not in model.metadata['system']:
                model.metadata['system']['annotationsSubsets'] = dict()
            model.metadata['system']['annotationsSubsets'][subset_name] = subset_annotation_filter_dict

        model.update(system_metadata=True)

    @staticmethod
    def delete_subset(model: entities.Model, subset_name: str):
        """
        Removes a subset from a model's metadata (both subsets and annotationsSubsets).

        :param dtlpy.entities.Model model: the model to which the subset should be added
        :param str subset_name: the name of the subset

        **Example**

        .. code-block:: python

            project.models.add_subset(model=model_entity, subset_name='train', subset_filter=dtlpy.Filters(field='dir', values='/train'))
            model_entity.metadata['system']['subsets']
                {'train': <dtlpy.entities.filters.Filters object at 0x1501dfe20>}
            project.models.delete_subset(model=model_entity, subset_name='train')
            model_entity.metadata['system']['subsets']
                {}

        """
        # Check if subset exists in subsets (for warning)
        if model.metadata.get("system", dict()).get("subsets", dict()).get(subset_name) is None:
            logger.error(f"Model system metadata incomplete, could not delete subset {subset_name}.")
        else:
            _ = model.metadata['system']['subsets'].pop(subset_name)

        # Remove from annotationsSubsets if it exists
        if model.metadata.get("system", dict()).get("annotationsSubsets", dict()).get(subset_name) is not None:
            _ = model.metadata['system']['annotationsSubsets'].pop(subset_name)

            model.update(system_metadata=True)

    def create(
            self,
            model_name: str,
            dataset_id: str = None,
            labels: list = None,
            ontology_id: str = None,
            description: str = None,
            model_artifacts: List[entities.Artifact] = None,
            project_id=None,
            tags: List[str] = None,
            package: entities.Package = None,
            configuration: dict = None,
            status: str = None,
            scope: entities.EntityScopeLevel = entities.EntityScopeLevel.PROJECT,
            version: str = '1.0.0',
            input_type=None,
            output_type=None,
            train_filter: entities.Filters = None,
            validation_filter: entities.Filters = None,
            annotations_train_filter: entities.Filters = None,
            annotations_validation_filter: entities.Filters = None,
            app: entities.App = None
    ) -> entities.Model:
        """
        Create a Model entity

        :param str model_name: name of the model
        :param str dataset_id: dataset id
        :param list labels: list of labels from ontology (must mach ontology id) can be a subset
        :param str ontology_id: ontology to connect to the model
        :param str description: description
        :param model_artifacts: optional list of dl.Artifact. Can be ItemArtifact, LocaArtifact or LinkArtifact
        :param str project_id: project that owns the model
        :param list tags: list of string tags
        :param package: optional - Package object
        :param dict configuration: optional - model configuration - dict
        :param str status: `str` of the optional values of
        :param str scope: the scope level of the model dl.EntityScopeLevel
        :param str version: version of the model
        :param str input_type: the file type the model expect as input (image, video, txt, etc)
        :param str output_type: dl.AnnotationType - the type of annotations the model produces (class, box segment, text, etc)
        :param dtlpy.entities.filters.Filters train_filter: Filters entity or a dictionary to define the items' scope in the specified dataset_id for the model train
        :param dtlpy.entities.filters.Filters validation_filter: Filters entity or a dictionary to define the items' scope in the specified dataset_id for the model validation
        :param dtlpy.entities.filters.Filters annotations_train_filter: Filters entity or a dictionary to define the annotations' scope in the specified dataset_id for the model train
        :param dtlpy.entities.filters.Filters annotations_validation_filter: Filters entity or a dictionary to define the annotations' scope in the specified dataset_id for the model validation
        :param dtlpy.entities.App app: App entity to connect the model to
        :return: Model Entity

        **Example**:

        .. code-block:: python

            project.models.create(model_name='model_name', dataset_id='dataset_id', labels=['label1', 'label2'], train_filter={filter: {$and: [{dir: "/10K short videos"}]},page: 0,pageSize: 1000,resource: "items"}})

        """

        if ontology_id is not None:
            # take labels from ontology
            ontologies = repositories.Ontologies(client_api=self._client_api)
            labels = [label.tag for label in ontologies.get(ontology_id=ontology_id).labels]

        if labels is None:
            # dont have to have labels. can use an empty list
            labels = list()

        if input_type is None:
            input_type = 'image'

        if output_type is None:
            output_type = entities.AnnotationType.CLASSIFICATION

        if package is None and self._package is None:
            raise exceptions.PlatformException('Must provide a Package or create from package.models')
        elif package is None:
            package = self._package

        # TODO need to remove the entire project id user interface - need to take it from dataset id (in BE)
        if project_id is None:
            if self._project is None:
                raise exceptions.PlatformException('Please provide project_id')
            project_id = self._project.id
        else:
            if project_id != self._project_id:
                if (isinstance(package, entities.Package) and not package.is_global) or \
                        (isinstance(package, entities.Dpk) and not package.scope != 'public'):
                    logger.warning(
                        "Note! you are specified project_id {!r} which is different from repository context: {!r}".format(
                            project_id, self._project_id))

        if model_artifacts is None:
            model_artifacts = []

        if not isinstance(model_artifacts, list):
            raise ValueError('`model_artifacts` must be a list of dl.Artifact entities')

        # create payload for request
        payload = {
            'packageId': package.id,
            'name': model_name,
            'projectId': project_id,
            'datasetId': dataset_id,
            'labels': labels,
            'artifacts': [artifact.to_json(as_artifact=True) for artifact in model_artifacts],
            'scope': scope,
            'version': version,
            'inputType': input_type,
            'outputType': output_type,
        }

        if app is not None:
            if not isinstance(package, entities.Dpk):
                raise ValueError('package must be a Dpk entity')
            if app.dpk_name != package.name or app.dpk_version != package.version:
                raise ValueError('App and package must be the same')
            component_name = None
            compute_config = None
            for model in package.components.models:
                if model['name'] == model_name:
                    component_name = model['name']
                    compute_config = model.get('computeConfigs', None)
                    break
            if component_name is None:
                raise ValueError('Model name not found in package')
            payload['app'] = {
                "id": app.id,
                "componentName": component_name,
                "dpkName": package.name,
                "dpkVersion": package.version
            }
            if compute_config is not None:
                payload['app']['computeConfig'] = compute_config

        if configuration is not None:
            payload['configuration'] = configuration

        if tags is not None:
            payload['tags'] = tags

        if description is not None:
            payload['description'] = description

        if status is not None:
            payload['status'] = status

        if train_filter or validation_filter or annotations_train_filter or annotations_validation_filter:
            metadata = Models._build_model_metadata(
                train_filter=train_filter,
                validation_filter=validation_filter,
                annotations_train_filter=annotations_train_filter,
                annotations_validation_filter=annotations_validation_filter
            )
            payload['metadata'] = metadata

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/ml/models',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        model = entities.Model.from_json(_json=response.json(),
                                         client_api=self._client_api,
                                         project=self._project,
                                         package=package)

        return model

    def clone(self,
              from_model: entities.Model,
              model_name: str,
              dataset: entities.Dataset = None,
              configuration: dict = None,
              status=None,
              scope=None,
              project_id: str = None,
              labels: list = None,
              description: str = None,
              tags: list = None,
              train_filter: entities.Filters = None,
              validation_filter: entities.Filters = None,
              annotations_train_filter: entities.Filters = None,
              annotations_validation_filter: entities.Filters = None,
              wait=True,
              ) -> entities.Model:
        """
        Clones and creates a new model out of existing one

        :param from_model: existing model to clone from
        :param str model_name: `str` new model name
        :param str dataset: dataset object for the cloned model
        :param dict configuration: `dict` (optional) if passed replaces the current configuration
        :param str status: `str` (optional) set the new status
        :param str scope: `str` (optional) set the new scope. default is "project"
        :param str project_id: `str` specify the project id to create the new model on (if other than the source model)
        :param list labels:  `list` of `str` - label of the model
        :param str description: `str` description of the new model
        :param list tags:  `list` of `str` - label of the model
        :param dtlpy.entities.filters.Filters train_filter: Filters entity or a dictionary to define the items' scope in the specified dataset_id for the model train
        :param dtlpy.entities.filters.Filters validation_filter: Filters entity or a dictionary to define the items' scope in the specified dataset_id for the model validation
        :param dtlpy.entities.filters.Filters annotations_train_filter: Filters entity or a dictionary to define the annotations' scope in the specified dataset_id for the model train
        :param dtlpy.entities.filters.Filters annotations_validation_filter: Filters entity or a dictionary to define the annotations' scope in the specified dataset_id for the model validation
        :param bool wait: `bool` wait for model to be ready
        :return: dl.Model which is a clone version of the existing model
        """
        from_json = {"name": model_name,
                     "packageId": from_model.package_id,
                     "configuration": from_model.configuration,
                     "outputType": from_model.output_type,
                     "inputType": from_model.input_type}
        if project_id is None:
            if dataset is not None:
                # take dataset project
                project_id = dataset.project.id
            else:
                # take model's project
                project_id = self.project.id
        from_json['projectId'] = project_id
        if dataset is not None:
            if labels is None:
                labels = list(dataset.labels_flat_dict.keys())
            from_json['datasetId'] = dataset.id
        if labels is not None:
            from_json['labels'] = labels
            # if there are new labels - pop the mapping from the original
            _ = from_json['configuration'].pop('id_to_label_map', None)
            _ = from_json['configuration'].pop('label_to_id_map', None)
        if configuration is not None:
            from_json['configuration'].update(configuration)
        if description is not None:
            from_json['description'] = description
        if tags is not None:
            from_json['tags'] = tags
        if scope is not None:
            from_json['scope'] = scope
        if status is not None:
            from_json['status'] = status

        metadata = Models._build_model_metadata(
            train_filter=train_filter,
            validation_filter=validation_filter,
            annotations_train_filter=annotations_train_filter,
            annotations_validation_filter=annotations_validation_filter,
            from_model=from_model
        )
        if metadata['system']['subsets'] or metadata['system']['annotationsSubsets']:
            from_json['metadata'] = metadata
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/ml/models/{}/clone'.format(from_model.id),
                                                         json_req=from_json)
        if not success:
            raise exceptions.PlatformException(response)
        new_model = entities.Model.from_json(_json=response.json(),
                                             client_api=self._client_api,
                                             project=self._project,
                                             package=from_model._package)
        if wait:
            new_model = self.wait_for_model_ready(model=new_model)
        return new_model

    def wait_for_model_ready(self, model: entities.Model):
        """
        Wait for model to be ready

        :param model: Model entity
        """
        sleep_time = MIN_INTERVAL
        while model.status == entities.ModelStatus.CLONING:
            model = self.get(model_id=model.id)
            time.sleep(sleep_time)
            sleep_time = min(sleep_time * BACKOFF_FACTOR, MAX_INTERVAL)
            time.sleep(sleep_time)
        return model

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/models".format(self.project.id))

    def open_in_web(self, model=None, model_id=None):
        """
        Open the model in web platform

        :param model: model entity
        :param str model_id: model id
        """
        if model is not None:
            model.open_in_web()
        elif model_id is not None:
            self._client_api._open_in_web(url=self.platform_url + '/' + str(model_id) + '/main')
        else:
            self._client_api._open_in_web(url=self.platform_url)

    def delete(self, model: entities.Model = None, model_name=None, model_id=None):
        """
        Delete Model object

        :param model: Model entity to delete
        :param str model_name: delete by model name
        :param str model_id: delete by model id
        :return: True
        :rtype: bool
        """
        # get id and name
        if model_id is None:
            if model is not None:
                model_id = model.id
            elif model_name is not None:
                model = self.get(model_name=model_name)
                model_id = model.id
            else:
                raise exceptions.PlatformException(error='400',
                                                   message='Must input at least one parameter to models.delete')

        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="/ml/models/{}".format(model_id)
        )

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return results
        return True

    def update(self,
               model: entities.Model,
               system_metadata: bool = False,
               reload_services: bool = True
               ) -> entities.Model:
        """
        Update Model changes to platform

        :param model: Model entity
        :param bool system_metadata: True, if you want to change metadata system
        :param bool reload_services: True, if you want to update services
        :return: Model entity
        """
        # payload
        payload = model.to_json()

        # url
        url_path = '/ml/models/{}'.format(model.id)
        query_params = {}
        if system_metadata:
            query_params['system'] = 'true'
        if reload_services is not None:
            query_params['reloadServices'] = 'true' if reload_services else 'false'

        if query_params:
            url_path += '?' + urlencode(query_params)

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Model.from_json(_json=response.json(),
                                        client_api=self._client_api,
                                        project=self._project,
                                        package=model._package)

    def train(self, model_id: str, service_config=None):
        """
        Train the model in the cloud. This will create a service and will run the adapter's train function as an execution

        :param model_id: id of the model to train
        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :return:
        """
        payload = dict()
        if service_config is not None:
            payload['serviceConfig'] = service_config
        success, response = self._client_api.gen_request(req_type="post",
                                                         path=f"/ml/models/{model_id}/train",
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        return entities.Execution.from_json(_json=response.json(),
                                            client_api=self._client_api,
                                            project=self._project)

    def evaluate(self, model_id: str, dataset_id: str, filters: entities.Filters = None, service_config=None):
        """
        Evaluate Model, provide data to evaluate the model on You can also provide specific config for the deployed service

        :param str model_id: Model id to predict
        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :param str dataset_id: ID of the dataset to evaluate
        :param entities.Filters filters: dl.Filter entity to run the predictions on
        :return:
        """

        payload = {'input': {'datasetId': dataset_id}}
        if service_config is not None:
            payload['config'] = {'serviceConfig': service_config}
        if filters is None:
            filters = entities.Filters()
        if filters is not None:
            payload['input']['datasetQuery'] = filters.prepare()
        success, response = self._client_api.gen_request(req_type="post",
                                                         path=f"/ml/models/{model_id}/evaluate",
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        return entities.Execution.from_json(_json=response.json(),
                                            client_api=self._client_api,
                                            project=self._project)

    def predict(self, model, item_ids=None, dataset_id=None,filters=None):
        """
        Run model prediction with items

        :param model: dl.Model entity to run the prediction.
        :param item_ids: a list of item id to run the prediction.
        :param dataset_id: a dataset id to run the prediction.
        :param filters_dict: dict of filters to run the prediction.
        :return:
        """
        if len(model.metadata['system'].get('deploy', {}).get('services', [])) == 0:
            # no services for model
            raise ValueError("Model doesnt have any associated services. Need to deploy before predicting")
        if item_ids is None and dataset_id is None:
            raise ValueError("Need to provide either item_ids or dataset_id")
        if filters is not None and dataset_id is None:
            raise ValueError("If filters are provided, dataset_id is mandatory.")
        payload_input = {}
        if item_ids is not None:
            payload_input['itemIds'] = item_ids
        if dataset_id is not None:
            payload_input['datasetId'] = dataset_id
        if filters is not None:
            payload_input['datasetQuery'] = filters.prepare()['filter']
        payload = {'input': payload_input,
                   'config': {'serviceId': model.metadata['system']['deploy']['services'][0]}}
        logger.debug(f"generate post request to predict with payload {payload}")
        success, response = self._client_api.gen_request(req_type="post",
                                                         path=f"/ml/models/{model.id}/predict",
                                                         json_req=payload)
        if not success:
            logger.error(f"failed to make API request /ml/models/{model.id}/predict with payload {payload} response {response}")
            raise exceptions.PlatformException(response)
        return entities.Execution.from_json(_json=response.json(),
                                            client_api=self._client_api,
                                            project=self._project)

    def embed(self, model, item_ids=None, dataset_id=None, filters=None):
        """
        Run model embed with items

        :param model: dl.Model entity to run the prediction.
        :param item_ids: a list of item id to run the embed.
        :param dataset_id: a dataset id to run the embed.
        :param filters_dict: dict of filters to run the embed.
        :return: Execution
        :rtype: dtlpy.entities.execution.Execution
        """
        if len(model.metadata['system'].get('deploy', {}).get('services', [])) == 0:
            # no services for model
            raise ValueError("Model doesnt have any associated services. Need to deploy before predicting")
        if item_ids is None and dataset_id is None:
            raise ValueError("Need to provide either item_ids or dataset_id")
        if filters is not None and dataset_id is None:
            raise ValueError("If filters are provided, dataset_id is mandatory.")
        payload_input = {}
        if item_ids is not None:
            payload_input['itemIds'] = item_ids
        if dataset_id is not None:
            payload_input['datasetId'] = dataset_id
        if filters is not None:
            payload_input['datasetQuery'] = filters.prepare()['filter']
        payload = {'input': payload_input,
                   'config': {'serviceId': model.metadata['system']['deploy']['services'][0]}}
        logger.debug(f"generate post request to embed with payload {payload}")
        success, response = self._client_api.gen_request(req_type="post",
                                                         path=f"/ml/models/{model.id}/embed",
                                                         json_req=payload)
        if not success:
            logger.error(f"failed to make API request /ml/models/{model.id}/embed with payload {payload} response {response}")
            raise exceptions.PlatformException(response)
        return entities.Execution.from_json(_json=response.json(),
                                            client_api=self._client_api,
                                            project=self._project)

    def embed_datasets(self, model, dataset_ids, attach_trigger=False):
        """
        Run model embed with datasets

        :param model: dl.Model entity to run the prediction.
        :param dataset_ids: a list of dataset id to run the embed.
        :param attach_trigger: bool, if True will activate the trigger
        :return:
        """
        if len(model.metadata['system'].get('deploy', {}).get('services', [])) == 0:
            # no services for model
            raise ValueError("Model doesnt have any associated services. Need to deploy before predicting")
        if dataset_ids is None:
            raise ValueError("Need to provide either dataset_id")
        payload = {'datasetIds': dataset_ids,
                   'config': {'serviceId': model.metadata['system']['deploy']['services'][0]},
                   'attachTrigger': attach_trigger
                   }

        success, response = self._client_api.gen_request(req_type="post",
                                                         path=f"/ml/models/{model.id}/embed/datasets",
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        command = entities.Command.from_json(_json=response.json(),
                                             client_api=self._client_api)
        command = command.wait()
        return command

    def deploy(self, model_id: str, service_config=None) -> entities.Service:
        """
        Deploy a trained model. This will create a service that will execute predictions

        :param model_id: id of the model to deploy
        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :return: dl.Service: the deployed service
        """
        payload = dict()
        if service_config is not None:
            payload['serviceConfig'] = service_config if not service_config.get("serviceConfig") else service_config.get("serviceConfig")
        success, response = self._client_api.gen_request(req_type="post",
                                                         path=f"/ml/models/{model_id}/deploy",
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)

        return entities.Service.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          project=self._project,
                                          package=self._package)


class Metrics:
    def __init__(self, client_api, model=None, model_id=None):
        self._client_api = client_api
        self._model_id = model_id
        self._model = model

    @property
    def model(self):
        return self._model

    def create(self, samples, dataset_id) -> bool:
        """
        Add Samples for model analytics and metrics

        :param samples: list of dl.PlotSample - must contain: model_id, figure, legend, x, y
        :param model_id: model id to save samples on
        :param dataset_id:
        :return: bool: True if success
        """
        if not isinstance(samples, list):
            samples = [samples]

        payload = list()
        for sample in samples:
            _json = sample.to_json()
            _json['modelId'] = self.model.id
            _json['datasetId'] = dataset_id
            payload.append(_json)
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/ml/metrics/publish',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return True

    def _list(self, filters: entities.Filters):
        # request
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path='/ml/metrics/query',
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Model]:
        jobs = [None for _ in range(len(response_items))]
        pool = self._client_api.thread_pools(pool_name='entity.create')

        # return triggers list
        for i_service, sample in enumerate(response_items):
            jobs[i_service] = pool.submit(entities.PlotSample,
                                          **{'x': sample.get('data', dict()).get('x', None),
                                             'y': sample.get('data', dict()).get('y', None),
                                             'legend': sample.get('legend', ''),
                                             'figure': sample.get('figure', '')})

        # get all results
        results = [j.result() for j in jobs]
        # return good jobs
        return miscellaneous.List(results)

    def list(self, filters=None) -> entities.PagedEntities:
        """
        List Samples for model analytics and metrics

        :param filters: dl.Filter query entity
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.METRICS)
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.METRICS:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.METRICS. Got: {!r}'.format(filters.resource))
        if self._model is not None:
            filters.add(field='modelId', values=self._model.id)
        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged



import traceback
import logging
import attr
import os

from .. import entities

logger = logging.getLogger(name='dtlpy')


class ArtifactType:
    ITEM = 'item'
    LOCAL = 'local'
    LINK = 'link'


@attr.s
class Artifact:
    @staticmethod
    def _protected_from_json(
            _json: dict,
            client_api,
            dataset=None,
            project=None,
            is_fetched: bool = True):
        """
        Same as from_json but with try-except to catch if error

        :param dict _json: platform json
        :param client_api: ApiClient entity
        :param dtlpy.entities.dataset.Dataset dataset: dataset entity
        :param dtlpy.entities.project.Project project: project entity
        :param bool is_fetched: is Entity fetched from Platform
        :return: status and the artifact object
        """
        try:
            artifact = Artifact.from_json(_json=_json,
                                          client_api=client_api,
                                          dataset=dataset,
                                          project=project,
                                          is_fetched=is_fetched)
            status = True
        except Exception:
            artifact = traceback.format_exc()
            status = False
        return status, artifact

    @classmethod
    def from_json(cls,
                  _json: dict,
                  client_api,
                  dataset=None,
                  project=None,
                  model=None,
                  is_fetched: bool = True):
        artifact_type = _json.get('type', None)
        if artifact_type is None:
            raise ValueError('Missing `type` in artifact.')
        if artifact_type in ['file', 'dir']:
            # used as item directly
            artifact = ItemArtifact.from_json(_json=_json,
                                              client_api=client_api,
                                              dataset=dataset,
                                              project=project,
                                              model=model,
                                              is_fetched=is_fetched
                                              )
        elif artifact_type in ['item']:
            # used as an item reference
            artifact = ItemArtifact.from_json(_json={'id': _json.get('itemId'),
                                                     'type': 'item'},
                                              client_api=client_api,
                                              dataset=dataset,
                                              project=project,
                                              model=model,
                                              is_fetched=False
                                              )
        elif artifact_type in ['local']:
            artifact = LocalArtifact.from_json(_json=_json)
        elif artifact_type in ['link']:
            artifact = LinkArtifact.from_json(_json=_json)
        else:
            raise ValueError('Unknown dl.Artifact `type`: {}'.format(artifact_type))
        return artifact


@attr.s
class ItemArtifact(entities.Item, Artifact):

    def to_json(self, as_artifact=False):
        """
        Return dict representation of the artifact
        Using a flag to keep old artifact behaviour (for packages, returning the item itself and not the artifact)

        :param as_artifact: it True, return the Artifact json, otherwise, return item for backward compatibility
        :return:
        """
        if as_artifact:
            _json = {'type': self.type,
                     'itemId': self.id}
        else:
            _json = super(ItemArtifact, self).to_json()
        return _json


@attr.s
class LocalArtifact(Artifact):
    type = 'local'
    local_path = attr.ib(repr=True)

    @classmethod
    def from_json(cls, _json: dict):
        """
        Build an Artifact entity object from a json

        :param dict _json: platform json
        :return: Artifact object
        :rtype: dtlpy.entities.artifact.Artifact
        """

        inst = cls(local_path=_json.get('localPath'))
        return inst

    def to_json(self, as_artifact=True):
        """
        Return dict representation of the artifact
        Using a flag to keep old artifact behaviour (for packages).
        This flag is only relevant for ItemArtifact

        :param as_artifact: not is use, only to keep same signature for all artifacts
        :return:
        """
        _json = {'type': self.type,
                 'localPath': self.local_path}
        return _json


@attr.s
class LinkArtifact(Artifact):
    type = 'link'
    url = attr.ib(repr=True)
    filename = attr.ib(repr=True)

    @classmethod
    def from_json(cls,
                  _json: dict,
                  is_fetched: bool = True):
        """
        Build an Artifact entity object from a json

        :param dict _json: platform json
        :rtype: dtlpy.entities.artifact.Artifact
        """
        url = _json.get('url')
        filename = _json.get('filename')
        if filename is None:
            filename = os.path.basename(url)
        inst = cls(url=url,
                   filename=filename)
        inst.is_fetched = is_fetched
        return inst

    def to_json(self, as_artifact=True):
        """
        Return dict representation of the artifact
        Using a flag to keep old artifact behaviour (for packages).
        This flag is only relevant for ItemArtifact

        :param as_artifact: not is use, only to keep same signature for all artifacts
        :return:
        """
        _json = {'type': self.type,
                 'url': self.url}
        return _json



DEFAULT_PACKAGE_ENTRY_POINT = 'main.py'
DEFAULT_PACKAGE_CLASS_NAME = 'ServiceRunner'
DEFAULT_PACKAGE_FUNCTION_NAME = 'run'
DEFAULT_PACKAGE_MODULE_NAME = 'default_module'
DEFAULT_PACKAGE_NAME = 'default_package'
